import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from collections.abc import MutableMapping
from typing import List, Dict, Union, Tuple
from pandas.api.types import is_list_like


INDEX1D = Union[pd.Index, List[str], List[bool], List[int], slice]
INDEX = Union[INDEX1D, Tuple[INDEX1D, INDEX1D]]
CINDEX = Union[slice, List[int]] # CINDEX: Converted Index

class MultiArrayView(MutableMapping):
    def __init__(self, multiarrays: Dict[str, np.ndarray], index: CINDEX):
        self.parent = multiarrays
        self.index = index
        self.multiarrays = {}

    def __getitem__(self, key: Union[str, "Ellipsis"]) -> Union[np.ndarray, Dict[str, np.ndarray]]:
        if key is Ellipsis:
            for key in self.parent:
                if key not in self.multiarrays:
                    self.multiarrays[key] = self.parent[key][self.index]
            return self.multiarrays

        if key not in self.multiarrays:
            if key in self.parent:
                self.multiarrays[key] = self.parent[key][self.index]
            else:
                raise ValueError(f"Key '{key}' does not exist!")
        return self.multiarrays[key]

    def __setitem__(self, key: str, value: object):
        raise ValueError("Cannot set key for MultiArrayView object!")

    def __delitem__(self, key: str):
        raise ValueError("Cannot delete key for MultiArrayView object!")

    def __iter__(self):
        return iter(self.parent)

    def __len__(self):
        return len(self.parent)

    def __repr__(self) -> str:
        return f"View of multiarrays with keys: {str(list(self.parent))[1:-1]}"


class MetadataView(MutableMapping):
    def __init__(self, metadata: dict = None):
        self.metadata = metadata if metadata is not None else dict()

    def __getitem__(self, key: str) -> Union[str, np.ndarray]:
        return self.metadata[key]

    def __setitem__(self, key: str, value: object):
        self.metadata[key] = value

    def __delitem__(self, key: str):
        del self.metadata[key]

    def __iter__(self):
        return iter(self.metadata)

    def __len__(self):
        return len(self.metadata)

    def __repr__(self) -> str:
        return f"View of metadata with keys: {str(list(self.metadata))[1:-1]}"



def _parse_index(parent: Union["UnimodalData", "UnimodalDataView"], index: INDEX) -> Tuple[CINDEX, CINDEX]:

    def _extract_indices_from_parent(parent: Union["UnimodalData", "UnimodalDataView"]) -> Tuple[pd.Index, pd.Index, CINDEX, CINDEX]:
        if hasattr(parent, "barcode_index"):
            return parent.parent.obs_names, parent.parent.var_names, parent.barcode_index, parent.feature_index
        else:
            return parent.obs_names, parent.var_names, None, None

    def _check_index_type(index_1d, index_name: str) -> INDEX1D:
        if isinstance(index_1d, slice):
            if not (isinstance(index_1d.start, int) or isinstance(index_1d.start, np.integer) or isinstance(index_1d.start, str) or (index_1d.start is None)):
                raise ValueError(f"Invalid slice.start '{index_1d.start}'. slice.start must be either None, integer or string!")
            if not (isinstance(index_1d.stop, int) or isinstance(index_1d.stop, np.integer) or isinstance(index_1d.stop, str) or (index_1d.stop is None)):
                raise ValueError(f"Invalid slice.stop '{index_1d.stop}'. slice.stop must be either None, integer or string!")
            if not (isinstance(index_1d.step, int) or isinstance(index_1d.step, np.integer) or (index_1d.step is None)):
                raise ValueError(f"Invalid slice.step '{index_1d.step}'. slice.step must be either None or integer!")
        elif not isinstance(index_1d, pd.Index):
            if not is_list_like(index_1d):
                index_1d = np.array([index_1d])
            elif not isinstance(index_1d, np.ndarray):
                index_1d = np.array(index_1d, copy = False)
            if index_1d.ndim != 1:
                raise ValueError(f"{index_name} index must be 1 dimension!")
            if index_1d.dtype.kind not in {'b', 'i', 'u', 'O', 'U'}:
                raise ValueError(f"Unknown {index_name} index dtype: {index_1d.dtype}!")
        return index_1d

    def _process_pd_index(base_idx: pd.Index, index_1d: pd.Index) -> List[int]:
        if index_1d.has_duplicates:
            index_1d = index_1d.drop_duplicates()
        indexer = base_idx.get_indexer(index_1d)
        indexer = indexer[indexer >= 0]
        return indexer

    def _parse_one_index(base_idx: pd.Index, index_1d: Union[INDEX1D, None], index_name: str, view_index: CINDEX = None) -> CINDEX:
        """ index_name: 'row' or 'column' """
        if view_index is not None:
            base_idx = base_idx[view_index]

        indexer = None

        if isinstance(index_1d, slice):
            step = 1 if index_1d.step is None else index_1d.step

            start = index_1d.start
            if isinstance(start, str):
                if start not in base_idx:
                    raise ValueError(f"Cannot locate slice.start '{start}' in {index_name} index!")
                start = base_idx.get_loc(start)
            elif start is None:
                start = 0
            else:
                if start < 0 or start >= base_idx.size:
                    raise ValueError(f"slice.start '{start}' is out of the boundary [0, {base_idx.size}) for {index_name} index!")

            stop = index_1d.stop
            if isinstance(stop, str):
                if stop not in base_idx:
                    raise ValueError(f"Cannot locate slice.stop '{stop}' in {index_name} index!")
                stop = base_idx.get_loc(stop) + np.sign(step) # if str , use [] instead of [)
            elif stop is None:
                stop = base_idx.size
            else:
                if stop - step < 0 or stop - step >= base_idx.size:
                    raise ValueError(f"slice.stop '{stop}' is out of the boundary [0, {base_idx.size}) for {index_name} index!")

            indexer = slice(start, stop, step)
        elif isinstance(index_1d, np.ndarray) and (index_1d.dtype.kind in {'b', 'i', 'u'}):
            assert index_1d.ndim == 1

            if index_1d.dtype.kind == 'b':
                if index_1d.size != base_idx.size:
                    raise ValueError(f"{index_name} index size does not match: actual size {base_idx.size}, input size {index_1d.size}!")
                indexer = np.where(index_1d)[0]
            elif index_1d.dtype.kind == 'i' or index_1d.dtype.kind == 'u':
                if np.any(index_1d < 0):
                    raise ValueError(f"Detect negative values in {index_name} index!")
                if np.any(index_1d >= base_idx.size):
                    raise ValueError(f"Detect values exceeding the largest valid position {base_idx.size - 1} in {index_name} index!")
                if np.unique(index_1d).size < index_1d.size:
                    raise ValueError(f"{index_name} index values are not unique!")
                indexer = index_1d
        else:
            if not isinstance(index_1d, pd.Index):
                assert isinstance(index_1d, np.ndarray) and index_1d.ndim == 1
                index_1d = pd.Index(index_1d)
            indexer = _process_pd_index(base_idx, index_1d)

        if view_index is not None:
            if isinstance(view_index, slice):
                if isinstance(indexer, slice):
                    indexer = slice(view_index.start + (view_index.step * indexer.start), view_index.start + (view_index.step * indexer.stop), view_index.step * indexer.step)
                else:
                    indexer = np.array(range(view_index.start, view_index.stop, view_index.step))[indexer]
            else:
                indexer = view_index[indexer]

        return indexer


    bidx = fidx = slice(None)
    base_bidx, base_fidx, view_bidx, view_fidx = _extract_indices_from_parent(parent)

    if isinstance(index, tuple):
        if len(index) > 2:
            raise ValueError(f"Index dimension {len(index)} exceeds 2!")
        elif len(index) == 1:
            bidx = _check_index_type(index[0], "row")
        else:
            bidx = _check_index_type(index[0], "row")
            fidx = _check_index_type(index[1], "column")
    else:
        bidx = _check_index_type(index, "row")

    return _parse_one_index(base_bidx, bidx, "row", view_index = view_bidx), _parse_one_index(base_fidx, fidx, "column", view_index = view_fidx)



class UnimodalDataView:
    def __init__(self, unidata: "UnimodalData", barcode_index: CINDEX, feature_index: CINDEX, cur_matrix: str, obj_name: str = "UnimodalData"):
        self.parent = unidata
        self.barcode_index = barcode_index
        self.feature_index = feature_index
        self._all_arrays = isinstance(barcode_index, np.ndarray) and isinstance(feature_index, np.ndarray)

        self.barcode_metadata = None
        self.feature_metadata = None
        self.matrices = {}
        self.barcode_multiarrays = MultiArrayView(unidata.barcode_multiarrays, barcode_index)
        self.feature_multiarrays = MultiArrayView(unidata.feature_multiarrays, feature_index)

        self.metadata = MetadataView()
        for key, value in unidata.metadata.items():
            # For views, only copy string objects
            if isinstance(value, str):
                self.metadata[key] = value

        def _get_size(index: CINDEX) -> int:
            if isinstance(index, slice):
                return (index.stop - index.start) // index.step
            return index.size

        self._cur_matrix = cur_matrix
        self._shape = (_get_size(barcode_index), _get_size(feature_index))
        self._obj_name = obj_name

    def __repr__(self, repr_dict: Dict[str, str] = None) -> str:
        repr_str = f"View of {self._obj_name} object with n_obs x n_vars = {self._shape[0]} x {self._shape[1]}"
        repr_str += f"\n    Genome: {self.parent.get_genome()}; Modality: {self.parent.get_modality()}"
        repr_str += f"\n    It contains {len(self.parent.matrices)} matrices: {str(list(self.parent.matrices))[1:-1]}"
        repr_str += f"\n    It currently binds to matrix '{self._cur_matrix}' as X\n" if len(self.parent.matrices) > 0 else "\n    It currently binds to no matrix\n"
        for key in ["obs", "var", "obsm", "varm"]:
            str_out = repr_dict[key] if (repr_dict is not None) and (key in repr_dict) else str(list(getattr(self.parent, key).keys()))[1:-1]
            repr_str += f"\n    {key}: {str_out}"
        repr_str += f"\n    uns: {str(list(self.metadata))[1:-1]}"
        return repr_str

    @property
    def obs(self) -> pd.DataFrame:
        if self.barcode_metadata is None:
            self.barcode_metadata = self.parent.barcode_metadata.iloc[self.barcode_index].copy(deep = False)
        return self.barcode_metadata

    @obs.setter
    def obs(self, obs: pd.DataFrame):
        raise ValueError("Cannot set obs for UnimodalDataView object!")

    @property
    def obs_names(self) -> pd.Index:
        return self.obs.index

    @obs_names.setter
    def obs_names(self, obs_names: pd.Index):
        raise ValueError("Cannot set obs_names for UnimodalDataView object!")

    @property
    def var(self) -> pd.DataFrame:
        if self.feature_metadata is None:
            self.feature_metadata = self.parent.feature_metadata.iloc[self.feature_index].copy(deep = False)
        return self.feature_metadata

    @var.setter
    def var(self, var: pd.DataFrame):
        raise ValueError("Cannot set var for UnimodalDataView object!")

    @property
    def var_names(self) -> pd.Index:
        return self.var.index

    @var_names.setter
    def var_names(self, var_names: pd.Index):
        raise ValueError("Cannot set var_names for UnimodalDataView object!")

    @property
    def X(self) -> Union[csr_matrix, None]:
        if self._cur_matrix not in self.matrices:
            X = self.parent.matrices.get(self._cur_matrix, None)
            if X is not None:
                self.matrices[self._cur_matrix] = X[self.barcode_index.reshape(-1, 1), self.feature_index] if self._all_arrays else X[self.barcode_index, self.feature_index]
        return self.matrices.get(self._cur_matrix, None)

    @X.setter
    def X(self, X: csr_matrix):
        raise ValueError("Cannot set X for UnimodalDataView object!")

    @property
    def obsm(self) -> Dict[str, np.ndarray]:
        return self.barcode_multiarrays

    @obsm.setter
    def obsm(self, obsm: Dict[str, np.ndarray]):
        raise ValueError("Cannot set obsm for UnimodalDataView object!")

    @property
    def varm(self) -> Dict[str, np.ndarray]:
        return self.feature_multiarrays

    @varm.setter
    def varm(self, varm: Dict[str, np.ndarray]):
        raise ValueError("Cannot set varm for UnimodalDataView object!")

    @property
    def uns(self) -> MetadataView:
        return self.metadata

    @uns.setter
    def uns(self, uns: MetadataView):
        raise ValueError("Cannot set uns for UnimodalDataView object!")

    @property
    def shape(self) -> Tuple[int, int]:
        return self._shape

    @shape.setter
    def shape(self, _shape: Tuple[int, int]):
        raise ValueError("Cannot set shape for UnimodalDataView object!")

    def current_matrix(self) -> str:
        return self._cur_matrix

    def select_matrix(self, key: str) -> None:
        if key not in self.parent.matrices:
            raise ValueError(f"Matrix key '{key}' does not exist!")
        self._cur_matrix = key

    def __getitem__(self, index: INDEX) -> "UnimodalDataView":
        barcode_index, feature_index = _parse_index(self, index)
        return UnimodalDataView(self.parent, barcode_index, feature_index, self._cur_matrix, obj_name = self._obj_name)

    def _copy_matrices(self) -> Dict[str, csr_matrix]:
        for key in self.parent.matrices:
            if key not in self.matrices:
                X = self.parent.matrices[key]
                self.matrices[key] = X[self.barcode_index.reshape(-1, 1), self.feature_index] if self._all_arrays else X[self.barcode_index, self.feature_index]
        return self.matrices

    def copy(self, deep: bool = True) -> "UnimodalData":
        """ If not deep, copy shallowly, which means that if contents of obsm, varm and matrices of the copied object is modified, the view object is also modified """
        return self.parent._copy_view(self, deep)



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
import h5py
from typing import Dict

import logging
logger = logging.getLogger(__name__)

from pegasusio import modalities, UnimodalData, CITESeqData, MultimodalData


def load_10x_h5_file_v2(h5_in: h5py.Group) -> MultimodalData:
    """Load 10x v2 format matrix from hdf5 file

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v2 formatted hdf5 file.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v2(h5_in)
    """
    data = MultimodalData()
    for genome in h5_in.keys():
        group = h5_in[genome]

        M, N = group["shape"][...]
        mat = csr_matrix(
            (
                group["data"][...],
                group["indices"][...],
                group["indptr"][...],
            ),
            shape=(N, M),
        )

        barcodes = group["barcodes"][...].astype(str)
        ids = group["genes"][...].astype(str)
        names = group["gene_names"][...].astype(str)

        unidata = UnimodalData({"barcodekey": barcodes},
            {"featurekey": names, "featureid": ids},
            {"X": mat},
            {"modality": "rna", "genome": genome}
        )
        unidata.separate_channels()

        data.add_data(unidata)

    return data


def load_10x_h5_file_v3(h5_in: h5py.Group) -> MultimodalData:
    """Load 10x v3 format matrix from hdf5 file, allowing detection of crispr and citeseq libraries

    Parameters
    ----------

    h5_in : h5py.Group
        An instance of h5py.Group class that is connected to a 10x v3 formatted hdf5 file.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file_v3(h5_in)
    """
    M, N = h5_in["matrix/shape"][...]
    bigmat = csr_matrix(
        (
            h5_in["matrix/data"][...],
            h5_in["matrix/indices"][...],
            h5_in["matrix/indptr"][...],
        ),
        shape=(N, M),
    )
    barcodes = h5_in["matrix/barcodes"][...].astype(str)
    df = pd.DataFrame(data = {"genome": h5_in["matrix/features/genome"][...].astype(str),
                              "feature_type": h5_in["matrix/features/feature_type"][...].astype(str),
                              "id": h5_in["matrix/features/id"][...].astype(str),
                              "name": h5_in["matrix/features/name"][...].astype(str)})

    genomes = list(df["genome"].unique())
    if "" in genomes:
        genomes.remove("")
    default_genome = genomes[0] if len(genomes) == 1 else None

    data = MultimodalData()
    gb = df.groupby(by = ["genome", "feature_type"])
    for name, group in gb:
        barcode_metadata = {"barcodekey": barcodes}
        feature_metadata = {"featurekey": group["name"].values, "featureid": group["id"].values}
        mat = bigmat[:, gb.groups[name]]

        genome = name[0] if (name[0] != "" or default_genome is None) else default_genome
        modality = "custom"
        if name[1] == "Gene Expression":
            modality = "rna"
        elif name[1] == "CRISPR Guide Capture":
            modality = "crispr"
        elif name[1] == "Antibody Capture":
            modality = "citeseq"

        if modality == "citeseq":
            unidata = CITESeqData(barcode_metadata, feature_metadata, {"raw.count": mat}, {"genome": genome, "modality": modality})
        else:
            unidata = UnimodalData(barcode_metadata, feature_metadata, {"X": mat}, {"genome": genome, "modality": modality})
        unidata.separate_channels()

        data.add_data(unidata)

    return data


def load_10x_h5_file(input_h5: str) -> MultimodalData:
    """Load 10x format matrix (either v2 or v3) from hdf5 file

    Parameters
    ----------

    input_h5 : `str`
        The matrix in 10x v2 or v3 hdf5 format.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pair per genome.

    Examples
    --------
    >>> io.load_10x_h5_file('example_10x.h5')
    """
    data = None
    with h5py.File(input_h5, 'r') as h5_in:
        load_file = load_10x_h5_file_v3 if "matrix" in h5_in.keys() else load_10x_h5_file_v2
        data = load_file(h5_in)

    return data


def load_loom_file(input_loom: str, genome: str = None, modality: str = None) -> MultimodalData:
    """Load count matrix from a LOOM file.

    Parameters
    ----------

    input_loom : `str`
        The LOOM file, containing the count matrix.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead. If not None and input loom contains genome attribute, the attribute will be overwritten.
    modality: `str`, optional (default None)
        Modality. If None, use "rna" instead. If not None and input loom contains modality attribute, the attribute will be overwritten.

    Returns
    -------

    A MultimodalData object containing a (genome, UmimodalData) pair.

    Examples
    --------
    >>> io.load_loom_file('example.loom', genome = 'GRCh38')
    """
    col_trans = {"CellID": "barcodekey", "obs_names": "barcodekey", "cell_names": "barcodekey"}
    row_trans = {
        "Gene": "featurekey", "var_names": "featurekey", "gene_names": "featurekey",
        "Accession": "featureid",  "gene_ids": "featureid", "ensembl_ids": "featureid",
    }

    import loompy
    with loompy.connect(input_loom) as ds:
        barcode_metadata = {}
        barcode_multiarrays = {}
        for key, arr in ds.col_attrs.items():
            key = col_trans.get(key, key)
            if arr.ndim == 1:
                barcode_metadata[key] = arr
            elif arr.ndim > 1:
                barcode_multiarrays[key] = arr
            else:
                raise ValueError(f"Detected column attribute '{key}' has ndim = {arr.ndim}!")

        feature_metadata = {}
        feature_multiarrays = {}
        for key, arr in ds.row_attrs.items():
            key = row_trans.get(key, key)
            if arr.ndim == 1:
                feature_metadata[key] = arr
            elif arr.ndim > 1:
                feature_multiarrays[key] = arr
            else:
                raise ValueError(f"Detected row attribute '{key}' has ndim = {arr.ndim}!")

        matrices = {}
        for key, mat in ds.layers.items():
            key = "X" if key == "" else key
            matrices[key] = mat.sparse().T.tocsr()

        metadata = dict(ds.attrs)
        if genome is not None:
            metadata["genome"] = genome
        elif "genome" not in metadata:
            metadata["genome"] = "unknown"

        if modality is not None:
            metadata["modality"] = modality
        elif "modality" not in metadata:
            if metadata.get("experiment_type", "none") in modalities:
                metadata["modality"] = metadata.pop("experiment_type")
            else:
                metadata["modality"] = "rna"

        unidata = UnimodalData(barcode_metadata, feature_metadata, matrices, metadata, barcode_multiarrays, feature_multiarrays)
        unidata.separate_channels()

    data = MultimodalData(unidata)
    return data


def write_loom_file(data: MultimodalData, output_file: str) -> None:
    """ Write a MultimodalData to loom file. Will assert data only contain one type of experiment.
    """
    keys = data.list_data()
    if len(keys) > 1:
        raise ValueError(f"Data contain multiple modalities: {','.join(keys)}!")
    data.select_data(keys[0])
    matrices = data.list_keys()
    assert "X" in matrices
    if len(matrices) == 0:
        raise ValueError("Could not write empty matrix to a loom file!")

    def _process_attrs(key_name: str, attrs: pd.DataFrame, attrs_multi: dict) -> Dict[str, object]:
        res_dict = {key_name: attrs.index.values}
        for key in attrs.columns:
            res_dict[key] = np.array(attrs[key].values)
        for key, value in attrs_multi.items():
            if value.ndim > 1: # value.ndim == 1 refers to np.recarray, which will not be written to a loom file.
                res_dict[key] = value if value.shape[1] > 1 else value[:, 0]
        return res_dict

    row_attrs = _process_attrs("Gene", data.var, data.varm)
    col_attrs = _process_attrs("CellID", data.obs, data.obsm)

    accession_key = "featureid" if "featureid" in row_attrs else ("gene_ids" if "gene_ids" in row_attrs else None)
    if accession_key is not None:
        row_attrs["Accession"] = row_attrs.pop(accession_key)

    layers = {}
    for matkey in matrices:
        layers["" if matkey == "X" else matkey] = data.get_matrix(matkey).T

    file_attrs = {}
    for key, value in data.uns.items():
        if isinstance(value, str):
            file_attrs[key] = value

    import loompy
    loompy.create(output_file, layers, row_attrs, col_attrs, file_attrs = file_attrs)

    logger.info(f"{output_file} is written.")



import gc
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix, hstack, vstack
from typing import List, Dict, Union, Set, Tuple, Optional

import logging
logger = logging.getLogger(__name__)

import anndata

from pegasusio import UnimodalData, VDJData, CITESeqData, CytoData, NanostringData
from pegasusio import calc_qc_filters, apply_qc_filters, DictWithDefault
from .views import INDEX, UnimodalDataView
from .datadict import MultiDataDict
from .vdj_data import VDJDataView


class MultimodalData:
    def __init__(self, unidata: Union[UnimodalData, anndata.AnnData, MultiDataDict] = None, genome: str = None, modality: str = None):
        self._selected = self._unidata = self._zarrobj = None

        if isinstance(unidata, MultiDataDict):
            self.data = unidata
        else:
            self.data = MultiDataDict()
            if unidata is not None:
                if isinstance(unidata, anndata.AnnData):
                    unidata = UnimodalData(unidata, genome = genome, modality = modality)
                self.add_data(unidata)


    def __repr__(self) -> str:
        repr_str = f"MultimodalData object with {len(self.data)} UnimodalData: {str(list(self.data))[1:-1]}"
        if self._selected is not None:
            repr_str += f"\n    It currently binds to {self._unidata.__class__.__name__} object {self._selected}\n\n"
            repr_str += self._unidata.__repr__()
        else:
            repr_str += "\n    It currently binds to no UnimodalData object"
        return repr_str


    def update(self, data: "MultimodalData") -> None:
        for key in data.data:
            if key in self.data:
                raise ValueError(f"Detected duplicated key '{key}'")
            self.data[key] = data.data[key]


    @property
    def obs(self) -> Union[pd.DataFrame, None]:
        return self._unidata.obs if self._unidata is not None else None

    @obs.setter
    def obs(self, obs: pd.DataFrame):
        assert self._unidata is not None
        self._unidata.obs = obs

    @property
    def obs_names(self) -> Union[pd.Index, None]:
        return self._unidata.obs_names if self._unidata is not None else None

    @obs_names.setter
    def obs_names(self, obs_names: pd.Index):
        assert self._unidata is not None
        self._unidata.obs_names = obs_names

    @property
    def var(self) -> Union[pd.DataFrame, None]:
        return self._unidata.var if self._unidata is not None else None

    @var.setter
    def var(self, var: pd.DataFrame):
        assert self._unidata is not None
        self._unidata.var = var

    @property
    def var_names(self) -> Union[pd.Index, None]:
        return self._unidata.var_names if self._unidata is not None else None

    @var_names.setter
    def var_names(self, var_names: pd.Index):
        assert self._unidata is not None
        self._unidata.var_names = var_names

    @property
    def X(self) -> Union[csr_matrix, np.ndarray, None]:
        return self._unidata.X if self._unidata is not None else None

    @X.setter
    def X(self, X: Union[csr_matrix, np.ndarray]):
        assert self._unidata is not None
        self._unidata.X = X

    @property
    def obsm(self) -> Union[Dict[str, np.ndarray], None]:
        return self._unidata.obsm if self._unidata is not None else None

    @obsm.setter
    def obsm(self, obsm: Dict[str, np.ndarray]):
        assert self._unidata is not None
        self._unidata.obsm = obsm

    @property
    def varm(self) -> Union[Dict[str, np.ndarray], None]:
        return self._unidata.varm if self._unidata is not None else None

    @varm.setter
    def varm(self, varm: Dict[str, np.ndarray]):
        assert self._unidata is not None
        self._unidata.varm = varm

    @property
    def uns(self) -> Union[dict, None]:
        return self._unidata.uns if self._unidata is not None else None

    @uns.setter
    def uns(self, uns: dict):
        assert self._unidata is not None
        self._unidata.uns = uns

    @property
    def shape(self) -> Tuple[int, int]:
        return self._unidata.shape if self._unidata is not None else None

    @shape.setter
    def shape(self, _shape: Tuple[int, int]):
        assert self._unidata is not None
        self._unidata.shape = _shape

    def as_float(self, matkey: str = None) -> None:
        """ Surrogate function to convert matrix to float """
        assert self._unidata is not None
        self._unidata.as_float(matkey)

    def list_keys(self, key_type: str = "matrix") -> List[str]:
        """ Surrogate function for UnimodalData, return available keys in metadata, key_type = barcode, feature, matrix, other
        """
        assert self._unidata is not None
        return self._unidata.list_keys(key_type)

    def current_matrix(self) -> str:
        """ Surrogate function for UnimodalData, return current matrix in current unimodal data
        """
        return self._unidata.current_matrix()

    def add_matrix(self, key: str, mat: csr_matrix) -> None:
        """ Surrogate function for UnimodalData, add a new matrix
        """
        assert self._unidata is not None
        self._unidata.add_matrix(key, mat)

    def select_matrix(self, key: str) -> None:
        """ Surrogate function for UnimodalData, select a matrix
        """
        assert self._unidata is not None
        self._unidata.select_matrix(key)

    def get_matrix(self, key: str) -> Union[csr_matrix, np.ndarray]:
        """ Surrogate function for UnimodalData, return a matrix indexed by key
        """
        assert self._unidata is not None
        return self._unidata.get_matrix(key)

    def get_modality(self) -> str:
        """ Surrogate function for UnimodalData, return modality, can be either 'rna', 'atac', 'tcr', 'bcr', 'crispr', 'hashing', 'citeseq' or 'cyto'.
        """
        assert self._unidata is not None
        return self._unidata.get_modality()

    def get_genome(self) -> str:
        """ Surrogate function for UnimodalData, returngenome
        """
        assert self._unidata is not None
        return self._unidata.get_genome()

    def _inplace_subset_obs(self, index: List[bool]) -> None:
        """ Surrogate function for UnimodalData, subset barcode_metadata inplace """
        assert self._unidata is not None
        self._unidata._inplace_subset_obs(index)

    def _inplace_subset_var(self, index: List[bool]) -> None:
        """ Surrogate function for UnimodalData, subset feature_metadata inplace """
        assert self._unidata is not None
        self._unidata._inplace_subset_var(index)

    def __getitem__(self, index: INDEX) -> Union[UnimodalDataView, VDJDataView]:
        """ Surrogate function for UnimodalData, [] operation """
        assert self._unidata is not None
        return self._unidata[index]


    def get_chain(self, chain: str) -> pd.DataFrame:
        """ Surrogate function for VDJData """
        assert self._unidata is not None and isinstance(self._unidata, VDJData)
        return self._unidata.get_chain(chain)

    def construct_clonotype(self, min_umis: int = 2) -> None:
        """ Surrogate function for VDJData """
        assert self._unidata is not None and isinstance(self._unidata, VDJData)
        self._unidata.construct_clonotype(min_umis=min_umis)

    def set_aside(self, params: List[str] = None) -> None:
        """ Surrogate function for CITESeqData and CytoData """
        assert self._unidata is not None and (isinstance(self._unidata, CITESeqData) or isinstance(self._unidata, CytoData))
        if params is None:
            self._unidata.set_aside()
        else:
            self._unidata.set_aside(params)

    def norm_hk(self, select: bool = True) -> None:
        """ Surrogate function for NanostringData"""
        assert self._unidata is not None and isinstance(self._unidata, NanostringData)
        self._unidata.norm_hk(select = select)

    def log_transform(self, select: bool = True) -> None:
        """ Surrogate function for NanostringData"""
        assert self._unidata is not None and isinstance(self._unidata, NanostringData)
        self._unidata.log_transform(select = select)

    def arcsinh_transform(self, cofactor: float = 5.0, jitter = False, random_state = 0, select: bool = True) -> None:
        """ Surrogate function for CITESeqData and CytoData"""
        assert self._unidata is not None and (isinstance(self._unidata, CITESeqData) or isinstance(self._unidata, CytoData))
        self._unidata.arcsinh_transform(cofactor = cofactor, jitter = jitter, random_state = random_state, select = select)



    def list_data(self) -> List[str]:
        return list(self.data)


    def add_data(self, unidata: UnimodalData) -> None:
        """ Add data, if _selected is not set, set as the first added dataset
        """
        key = unidata.get_uid()
        assert key is not None
        if key in self.data:
            raise ValueError(f"Key '{key}' already exists!")
        self.data[key] = unidata
        if self._selected is None:
            self._selected = key
            self._unidata = unidata


    def select_data(self, key: str) -> None:
        if key not in self.data:
            raise ValueError(f"Key '{key}' does not exist!")
        self._selected = key
        self._unidata = self.data[self._selected]


    def current_key(self) -> str:
        return self._selected


    def current_data(self) -> UnimodalData:
        return self._unidata


    def get_data(self, key: str = None, genome: str = None, modality: str = None, keep_list: bool = False) -> Union[UnimodalData, List[UnimodalData]]:
        """ get UnimodalData or list of UnimodalData based on MultimodalData key or genome or modality; accept negation '~' before genome or modality
            keep_list = True will return a list even with one data point.
        """
        if key is not None:
            if key not in self.data:
                raise ValueError(f"Key '{key}' does not exist!")
            return self.data[key]

        data_arr = []
        negation = False
        if genome is not None:
            if genome[0] == "~":
                negation = True
                genome = genome[1:]

            for unidata in self.data.values():
                cur_genome = unidata.get_genome()
                if ((not negation) and (cur_genome == genome)) or (negation and (cur_genome != genome)):
                    data_arr.append(unidata)

            if len(data_arr) == 0 and not keep_list:
                raise ValueError(f"No UnimodalData {'without' if negation else 'with'} genome '{genome}'!")
        else:
            if modality is None:
                raise ValueError("Either key or genome or modality needs to be set!")

            if modality[0] == "~":
                negation = True
                modality = modality[1:]

            for unidata in self.data.values():
                cur_modality = unidata.get_modality()
                if ((not negation) and (cur_modality == modality)) or (negation and (cur_modality != modality)):
                    data_arr.append(unidata)

            if len(data_arr) == 0 and not keep_list:
                raise ValueError(f"No UnimodalData {'without' if negation else 'with'} modality '{modality}'!")

        results = None
        if len(data_arr) == 1 and not keep_list:
            results = data_arr[0]
        else:
            results = data_arr

        return results


    def drop_data(self, key: str) -> UnimodalData:
        if key not in self.data:
            raise ValueError("Key {} does not exist!".format(key))
        return self.data.pop(key)


    def filter_data(self,
        select_singlets: Optional[bool] = False,
        remap_string: Optional[str] = None,
        subset_string: Optional[str] = None,
        min_genes: Optional[int] = None,
        max_genes: Optional[int] = None,
        min_umis: Optional[int] = None,
        max_umis: Optional[int] = None,
        mito_prefix: Optional[str] = None,
        percent_mito: Optional[float] = None,
        focus_list: Optional[Union[List[str], str]] = None,
        cache_passqc: Optional[bool] = False,
        uns_white_list: Optional[str] = None,
    ) -> None:
        """
        Filter each "rna" modality UnimodalData in the focus_list separately using the set filtration parameters. Then for all other UnimodalData objects, select only barcodes that are in the union of selected barcodes from previously filtered UnimodalData objects.
        If focus_list is None, focus_list = [self._selected]

        Parameters
        ----------
        select_singlets: ``bool``, optional, default ``False``
            If select only singlets.
        remap_string: ``str``, optional, default ``None``
            Remap singlet names using <remap_string>, where <remap_string> takes the format "new_name_i:old_name_1,old_name_2;new_name_ii:old_name_3;...". For example, if we hashed 5 libraries from 3 samples sample1_lib1, sample1_lib2, sample2_lib1, sample2_lib2 and sample3, we can remap them to 3 samples using this string: "sample1:sample1_lib1,sample1_lib2;sample2:sample2_lib1,sample2_lib2". In this way, the new singlet names will be in metadata field with key 'assignment', while the old names will be kept in metadata field with key 'assignment.orig'.
        subset_string: ``str``, optional, default ``None``
            If select singlets, only select singlets in the <subset_string>, which takes the format "name1,name2,...". Note that if --remap-singlets is specified, subsetting happens after remapping. For example, we can only select singlets from sampe 1 and 3 using "sample1,sample3".
        min_genes: ``int``, optional, default: None
            Only keep cells with at least ``min_genes`` genes.
        max_genes: ``int``, optional, default: None
            Only keep cells with less than ``max_genes`` genes.
        min_umis: ``int``, optional, default: None
            Only keep cells with at least ``min_umis`` UMIs.
        max_umis: ``int``, optional, default: None
            Only keep cells with less than ``max_umis`` UMIs.
        mito_prefix: ``str``, optional, default: None
            Prefix for mitochondrial genes. For example, GRCh38:MT-,mm10:mt-.
        percent_mito: ``float``, optional, default: None
            Only keep cells with percent mitochondrial genes less than ``percent_mito`` % of total counts. Only when both mito_prefix and percent_mito set, the mitochondrial filter will be triggered.
        focus_list: ``List[str]`` or ``str``, optional, default None
            Filter each UnimodalData with key in focus_list (and modality is 'rna') separately using the filtration criteria. If only one focus is provided, focus_list can be a string instead of a list.
        cache_passqc: ``bool``, optional, default: ``False``
            If True and "passed_qc" is in a UnimodalData object's obs field, use the cached "passed_qc" instead of recalculating it using 'calc_qc_filters'.
        uns_white_list: ``str``, optional, default: ``None``
            If not None, this is string representing a comma-separated list of uns keywords that should not be removed. In addition, 'genome' and 'modality' will not be removed. All other keys in uns will be removed. If None, do not remove any keyword from uns.
        """
        selected_barcodes = None

        if focus_list is None:
            focus_list = [self._selected]
        elif isinstance(focus_list, str):
            focus_list = [focus_list]
        focus_set = set(focus_list)

        unselected = []
        mito_dict = DictWithDefault(mito_prefix)
        for key, unidata in self.data.items():
            if (key in focus_set) and (unidata.get_modality() == "rna"):
                if ("passed_qc" not in unidata.obs) or (not cache_passqc):
                    calc_qc_filters(unidata,
                        select_singlets = select_singlets,
                        remap_string = remap_string,
                        subset_string = subset_string,
                        min_genes = min_genes,
                        max_genes = max_genes,
                        min_umis = min_umis,
                        max_umis = max_umis,
                        mito_prefix = mito_dict.get(unidata.get_genome()),
                        percent_mito = percent_mito)
                apply_qc_filters(unidata, uns_white_list=uns_white_list)
                selected_barcodes = unidata.obs_names if selected_barcodes is None else selected_barcodes.union(unidata.obs_names)
            else:
                unselected.append(unidata)

        if (selected_barcodes is not None) and len(unselected) > 0:
            for unidata in unselected:
                selected = unidata.obs_names.isin(selected_barcodes)
                prior_n = unidata.shape[0]
                unidata._inplace_subset_obs(selected)
                logger.info(f"After filtration, {unidata.shape[0]} out of {prior_n} cell barcodes are kept in UnimodalData object {unidata.get_uid()}.")


    def concat_data(self, modality: str = "rna"):
        """ Used for raw data, Ignore multiarrays and only consider one matrix per unidata """
        genomes = []
        unidata_arr = []

        for key in list(self.data):
            unidata = self.data.pop(key)
            assert unidata.get_modality() == modality
            genomes.append(unidata.get_genome())
            unidata_arr.append(unidata)

        unikey = None
        if len(genomes) == 1:
            unikey = unidata_arr[0].get_uid()
            self.data[unikey] = unidata_arr[0]
        else:
            genome = ",".join(genomes)
            feature_metadata = pd.concat([unidata.feature_metadata for unidata in unidata_arr], axis = 0)
            feature_metadata.reset_index(inplace = True)
            feature_metadata.fillna(value = "N/A", inplace = True)
            X = hstack([unidata.matrices["X"] for unidata in unidata_arr], format = "csr")
            unidata = UnimodalData(unidata_arr[0].barcode_metadata, feature_metadata, {"X": X}, {"genome": genome, "modality": "rna"})
            unikey = unidata.get_uid()
            self.data[unikey] = unidata
            del unidata_arr
            gc.collect()

        self._selected = unikey
        self._unidata = self.data[unikey]


    def subset_data(self, data_subset: Set[str] = None, genome_subset: Set[str] = None, modality_subset: Set[str] = None) -> None:
        """ Only keep data that are in data_subset or genome_subset or modality_subset
        """
        if data_subset is not None:
            for key in self.list_data():
                if key not in data_subset:
                    del self.data[key]
        elif genome_subset is not None:
            for key in self.list_data():
                if self.data[key].get_genome() not in genome_subset:
                    del self.data[key]
        elif modality_subset is not None:
            for key in self.list_data():
                if self.data[key].get_modality() not in modality_subset:
                    del self.data[key]

        if len(self.list_data()) == 0:
            self._selected = None
            self._unidata = None
        elif self._selected not in self.list_data():
            self._selected = self.list_data()[0]
            self._unidata = self.data[self._selected]


    def scan_black_list(self, black_list: Set[str] = None):
        """ Remove unwanted keys in the black list
            Note: black_list might be changed.
        """
        if black_list is None:
            return None

        def _check_reserved_keyword(black_list: Set[str], keyword: str):
            if keyword in black_list:
                logger.warning("Removed reserved keyword '{}' from black list.".format(keyword))
                black_list.remove(keyword)

        _check_reserved_keyword(black_list, "genome")
        _check_reserved_keyword(black_list, "modality")

        for key in self.data:
            self.data[key].scan_black_list(black_list)


    def to_anndata(self) -> anndata.AnnData:
        """ Convert current data to an anndata object
        """
        if self._unidata is None:
            raise ValueError("Please first select a unimodal data to convert!")
        return self._unidata.to_anndata()


    def copy(self) -> "MultimodalData":
        from copy import deepcopy
        new_data = MultimodalData(deepcopy(self.data))
        new_data._selected = self._selected
        new_data._zarrobj = None # Should not copy _zarrobj
        if new_data._selected is not None:
            new_data._unidata = new_data.data[new_data._selected]
        return new_data


    def __deepcopy__(self, memo):
        return self.copy()


    def kick_start(self):
        """ Begin to track changes in self.data """
        self.data.kick_start(self._selected)


    def write_back(self):
        """ Write back changes and clear dirty bits
        """
        assert self._zarrobj is not None
        if self.data.is_dirty():
            self._zarrobj.write_multimodal_data(self, overwrite = False)
            self.data.clear_dirty(self._selected)


    def to_zip(self):
        """ If data is backed as Zarr directory, convert it to zarr.zip
        """
        assert self._zarrobj is not None
        self._zarrobj._to_zip()


    def _update_barcode_metadata_info(
        self, row: pd.Series, attributes: Set[str], append_sample_name: bool
    ) -> None:
        for unidata in self.data.values():
            unidata._update_barcode_metadata_info(row, attributes, append_sample_name)


    def _update_genome(self, genome_dict: Dict[str, str]) -> None:
        for key in self.list_data():
            genome = self.data[key].get_genome()
            if genome in genome_dict:
                unidata = self.data.pop(key)
                unidata.uns["genome"] = genome_dict[genome]
                self.data[unidata.get_uid()] = unidata


    def _propogate_genome(self) -> None:
        genomes = set()
        unknowns = []
        for key in self.data:
            genome = self.data[key].get_genome()
            if genome == "unknown":
                unknowns.append(key)
            else:
                genomes.add(genome)

        if len(genomes) == 1 and len(unknowns) > 0:
            genome = list(genomes)[0]
            for key in unknowns:
                unidata = self.data.pop(key)
                unidata.uns["genome"] = genome
                self.data[unidata.get_uid()] = unidata


    def _convert_attributes_to_categorical(self, attributes: Set[str]) -> None:
        for unidata in self.data.values():
            unidata._convert_attributes_to_categorical(attributes)


    def _clean_tmp(self) -> dict:
        _tmp_multi = {}
        for key, unidata in self.data.items():
            _tmp_dict = unidata._clean_tmp()
            if _tmp_dict is not None:
                _tmp_multi[key] = _tmp_dict
        return _tmp_multi

    def _addback_tmp(self, _tmp_multi) -> None:
        for key, _tmp_dict in _tmp_multi.items():
            self.data[key]._addback_tmp(_tmp_dict)



# cython: language_level=3

import numpy as np
cimport cython

ctypedef unsigned char uint8


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple split_barcode_channel(str[:] arr):
    cdef Py_ssize_t size = arr.size, i

    barcodes = np.empty(size, dtype = np.object)
    channels = np.empty(size, dtype = np.object)

    cdef str[:] bview = barcodes, cview = channels
    cdef list res

    for i in range(size):
        res = arr[i].rsplit(sep = '-', maxsplit = 1)
        bview[i] = res[0]
        cview[i] = res[1]

    return (barcodes, channels)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple convert_10x_vdj_to_vdjdata(str[:] tokens, int[:, :] mats_int, str[:, :] mats_str, dict fid2pos, int n_barcodes, int n_contigs):
    cdef Py_ssize_t i, j
    cdef str chain, strval
    cdef int bpos = -1, fpos, num, delta, value
    cdef dict fmap = {}, strmap
    cdef list tracing = []
    cdef int n_features = len(fid2pos), nint = mats_int.shape[1], nstr = mats_str.shape[1], nmat = nint + nstr - 2

    barcodes = np.empty(n_barcodes, dtype = np.object)
    is_cell = np.zeros(n_barcodes, dtype = np.bool_)
    matrices = np.zeros((nmat, n_barcodes, n_features), dtype = np.int32)
    cdef list strarrs = []

    cdef str[:] bview = barcodes
    cdef uint8[:] icview = is_cell
    cdef int[:, :, :] matview = matrices
    cdef str[:] strview


    for i in range(nstr - 1):
        tracing.append({"None": 0, "": 0})

    for i in range(tokens.size):
        if bpos < 0 or bview[bpos] != tokens[i]:
            bpos += 1
            bview[bpos] = tokens[i]
            icview[bpos] = mats_int[i, nint - 1]
            fmap.clear()

        chain = mats_str[i, nstr - 1]
        num = fmap.get(chain, 0)
        if num >= n_contigs:
            raise ValueError(f"There are more than {n_contigs} productive contigs with chain type {chain}!")
        fpos = fid2pos[chain + (str(num + 1) if num > 0 else "")]
        for j in range(nint - 1):
            matview[j, bpos, fpos] = mats_int[i, j]
        delta = nint - 1
        for j in range(nstr - 1):
            strmap = tracing[j]
            strval = mats_str[i, j]
            value = strmap.get(strval, -1)
            if value < 0:
                value = len(strmap) - 1
                strmap[strval] = value
            matview[j + delta, bpos, fpos] = value
        fmap[chain] = num + 1

    for i in range(nstr - 1):
        del tracing[i]["None"]
        strmap = tracing[i]
        strarr = np.empty(len(strmap), dtype = np.object)
        strview = strarr
        for key, value in strmap.items():
            strview[value] = key
        strarrs.append(strarr)

    return barcodes, is_cell, matrices, strarrs



import os
import gzip
import anndata
from typing import Tuple, Set, Union, List
from pandas.api.types import is_list_like

import logging
logger = logging.getLogger(__name__)

from pegasusio import timer
from pegasusio import UnimodalData, MultimodalData


from .hdf5_utils import load_10x_h5_file, load_loom_file, write_loom_file
from .text_utils import load_mtx_file, write_mtx_file, load_csv_file, write_scp_file
from .zarr_utils import ZarrFile
from .vdj_utils import load_10x_vdj_file
from .cyto_utils import load_fcs_file
from .nanostring_utils import load_nanostring_files


def infer_file_type(input_file: Union[str, List[str]]) -> Tuple[str, str, str]:
    """ Infer file format from input_file name

    This function infer file type by inspecting the file name.

    Parameters
    ----------

    input_file : `str` or `List[str]`
        Input file name.

    Returns
    -------
    `str`
        File type, choosing from 'zarr', 'h5ad', 'loom', '10x', 'mtx', 'csv', 'tsv', 'fcs' or 'nanostring'.
    `str`
        The path covering all input files. Most time this is the same as input_file. But for HCA mtx and csv, this should be parent directory.
    `str`
        Type of the path, either 'file' or 'directory'.

    Note: The last two `str`s are mainly used for transfer to cloud
    """
    file_type = None
    copy_path = input_file
    copy_type = "file"

    if is_list_like(input_file):
        # now it can only be nanostring
        file_type = "nanostring"
        return file_type, copy_path, copy_type

    if input_file.endswith(".zarr") or input_file.endswith(".zarr.zip"):
        file_type = "zarr"
    elif input_file.endswith(".h5ad"):
        file_type = "h5ad"
    elif input_file.endswith(".loom"):
        file_type = "loom"
    elif input_file.endswith(".h5"):
        file_type = "10x"
    elif input_file.endswith(".fcs"):
        file_type = "fcs"
    elif (
        input_file.endswith(".mtx")
        or input_file.endswith(".mtx.gz")
        or os.path.splitext(input_file)[1] == ""
    ):
        file_type = "mtx"
        if os.path.splitext(input_file)[1] != "":
            copy_path = os.path.dirname(input_file)
        copy_type = "directory"
    elif input_file.endswith(".csv") or input_file.endswith(".csv.gz"):
        file_type = "csv"
        if os.path.basename(input_file) == "expression.csv":
            copy_path = os.path.dirname(input_file)
            copy_type = "directory"
    elif input_file.endswith(".txt") or input_file.endswith(".tsv") or input_file.endswith(
        ".txt.gz") or input_file.endswith(".tsv.gz"):
        file_type = "tsv"
    else:
        raise ValueError(f"Unrecognized file type for file {input_file}!")

    return file_type, copy_path, copy_type


def is_vdj_file(input_csv: str, file_type: str) -> bool:
    if file_type != "csv":
        return False
    with (gzip.open(input_csv, "rt") if input_csv.endswith(".gz") else open(input_csv)) as fin:
        line = next(fin)
        return line.find(",chain,v_gene,d_gene,j_gene,c_gene,") >= 0


@timer(logger=logger)
def read_input(
    input_file: str,
    file_type: str = None,
    mode: str = "r",
    genome: str = None,
    modality: str = None,
    black_list: Set[str] = None,
    select_data: Set[str] = None,
    select_genome: Set[str] = None,
    select_modality: Set[str] = None,
) -> MultimodalData:
    """Load data into memory.

    This function is used to load input data into memory. Inputs can be in 'zarr', 'h5ad', 'loom', '10x', 'mtx', 'csv', 'tsv', 'fcs' (for flow/mass cytometry data) or 'nanostring' (Nanostring GeoMx spatial data) formats.

    Parameters
    ----------

    input_file : `str`
        Input file name.
    file_type : `str`, optional (default: None)
        File type, choosing from 'zarr', 'h5ad', 'loom', '10x', 'mtx', 'csv', 'tsv', 'fcs' (for flow/mass cytometry data) or 'nanostring'. If None, inferred from input_file.
    mode: `str`, optional (default: 'r')
        File open mode, options are 'r' or 'a'. If mode == 'a', file_type must be zarr and ngene/select_singlets cannot be set.
    genome : `str`, optional (default: None)
        For formats like loom, mtx, dge, csv and tsv, genome is used to provide genome name. In this case if genome is None, except mtx format, "unknown" is used as the genome name instead.
    modality : `str`, optional (default: None)
        Default modality, choosing from 'rna', 'atac', 'tcr', 'bcr', 'crispr', 'hashing', 'citeseq', 'cyto' (flow cytometry / mass cytometry) or 'nanostring'. If None, use 'rna' as default.
    black_list : `Set[str]`, optional (default: None)
        Attributes in black list will be poped out.
    select_data: `Set[str]`, optional (default: None)
        Only select data with keys in select_data. Select_data, select_genome and select_modality are mutually exclusive.
    select_genome: `Set[str]`, optional (default: None)
        Only select data with genomes in select_genome. Select_data, select_genome and select_modality are mutually exclusive.
    select_modality: `Set[str]`, optional (default: None)
        Only select data with modalities in select_modality. Select_data, select_genome and select_modality are mutually exclusive.

    Returns
    -------

    A MultimodalData object.

    Examples
    --------
    >>> data = io.read_input('example_10x.h5')
    >>> data = io.read_input('example.h5ad')
    >>> data = io.read_input('example_ADT.csv', genome = 'hashing_HTO', modality = 'hashing')
    """
    
    if is_list_like(input_file):
        input_file = [os.path.expanduser(os.path.expandvars(x)) for x in input_file]
    else:
        input_file = os.path.expanduser(os.path.expandvars(input_file))

    if file_type is None:
        file_type, _, _ = infer_file_type(input_file)

    if mode == "a":
        if file_type != "zarr":
            raise ValueError("Only Zarr file can have mode 'a'!")
        zf = ZarrFile(input_file, mode = mode)
        data = zf.read_multimodal_data(attach_zarrobj = True)
    else:
        if file_type == "zarr":
            zf = ZarrFile(input_file)
            data = zf.read_multimodal_data()
        elif file_type == "h5ad":
            data = MultimodalData(anndata.read_h5ad(input_file), genome = genome, modality = modality)
        elif file_type == "loom":
            data = load_loom_file(input_file, genome = genome, modality = modality)
        elif file_type == "10x":
            data = load_10x_h5_file(input_file)
        elif file_type == "fcs":
            data = load_fcs_file(input_file, genome = genome)
        elif file_type == "nanostring":
            input_matrix = input_file[0]
            segment_file = input_file[1]
            annotation_file = input_file[2] if len(input_file) > 2 else None
            data = load_nanostring_files(input_matrix, segment_file, annotation_file = annotation_file, genome = genome)
        elif file_type == "mtx":
            data = load_mtx_file(input_file, genome = genome, modality = modality)
        else:
            assert file_type == "csv" or file_type == "tsv"
            if is_vdj_file(input_file, file_type):
                data = load_10x_vdj_file(input_file, genome = genome, modality = modality)
            else:
                data = load_csv_file(input_file, sep = "," if file_type == "csv" else "\t", genome = genome, modality = modality)

    data.subset_data(select_data, select_genome, select_modality)
    data.kick_start()
    data.scan_black_list(black_list)

    logger.info(f"{file_type} file '{input_file}' is loaded.")

    return data


@timer(logger=logger)
def write_output(
    data: Union[MultimodalData, UnimodalData],
    output_file: str,
    file_type: str = None,
    is_sparse: bool = True,
    precision: int = 2
) -> None:
    """ Write data back to disk.

    This function is used to write data back to disk.

    Parameters
    ----------

    data : MutimodalData
        data to write back.
    output_file : `str`
        output file name. Note that for mtx files, output_file specifies a directory. For scp format, file_type must be specified. 
    file_type : `str`, optional (default: None)
        File type can be 'zarr' (as folder), 'zarr.zip' (as a ZIP file), 'h5ad', 'loom', 'mtx' or 'scp'. If file_type is None, it will be inferred based on output_file.
    is_sparse : `bool`, optional (default: True)
        Only used for writing out SCP-compatible files, if write expression as a sparse matrix.
    precision : `int`, optional (default: 2)
        Precision after decimal point for values in mtx and scp expression matrix.

    Returns
    -------
    `None`

    Examples
    --------
    >>> io.write_output(data, 'test.zarr')
    """
    if isinstance(data, UnimodalData):
        data = MultimodalData(data)

    output_file = os.path.expanduser(os.path.expandvars(output_file))

    def _infer_output_file_type(output_File: str) -> str:
        if output_file.endswith(".zarr"):
            return "zarr"
        elif output_file.endswith(".zarr.zip"):
            return "zarr.zip"
        elif output_file.endswith(".h5ad"):
            return "h5ad"
        elif output_file.endswith(".loom"):
            return "loom"
        else:
            name, sep, suf = output_file.rpartition(".")
            return "mtx" if sep == "" else suf

    file_type = _infer_output_file_type(output_file) if file_type is None else file_type
    if file_type not in {"zarr", "zarr.zip", "h5ad", "loom", "mtx", "scp"}:
        raise ValueError(f"Unsupported output file type '{file_type}'!")

    _tmp_multi = data._clean_tmp() # for each unidata, remove uns keys starting with '_tmp' and store these values to _tmp_multi

    if file_type.startswith("zarr"):
        zf = ZarrFile(output_file, mode = "w", storage_type = "ZipStore" if file_type == "zarr.zip" else None)
        zf.write_multimodal_data(data)
        del zf
    elif file_type == "h5ad":
        data.to_anndata().write(output_file, compression="gzip")
    elif file_type == "loom":
        write_loom_file(data, output_file)
    elif file_type == "mtx":
        write_mtx_file(data, output_file, precision = precision)
    else:
        assert file_type == "scp"
        write_scp_file(data, output_file, is_sparse = is_sparse, precision = precision)

    data._addback_tmp(_tmp_multi)
    logger.info(f"{file_type} file '{output_file}' is written.")



import os
import re
import shlex
import numpy as np
import pandas as pd
from scipy.io import mmread, mmwrite
from scipy.sparse import csr_matrix
import tempfile
import subprocess
from typing import List, Dict, Tuple, Union

import logging
logger = logging.getLogger(__name__)

from pegasusio import UnimodalData, CITESeqData, MultimodalData


def _enumerate_files(path: str, parts: List[str], repl_list1: List[str], repl_list2: List[str] = None) -> str:
    """ Enumerate all possible file names """
    if len(parts) <= 2:
        for token in repl_list1:
            parts[-1] = token
            candidate = os.path.join(path, ''.join(parts))
            if os.path.isfile(candidate):
                return candidate
    else:
        assert len(parts) == 4
        for p2 in repl_list1:
            parts[1] = p2
            for p4 in repl_list2:
                parts[3] = p4
                candidate = os.path.join(path, ''.join(parts))
                if os.path.isfile(candidate):
                    return candidate
    return None


def _locate_barcode_and_feature_files(path: str, fname: str) -> Tuple[str, str]:
    """ Locate barcode and feature files (with path) based on mtx file name (no suffix)
    """
    barcode_file = feature_file = None
    if fname == "matrix":
        barcode_file = _enumerate_files(path, [''], ["cells.tsv.gz", "cells.tsv", "barcodes.tsv.gz", "barcodes.tsv"])
        feature_file = _enumerate_files(path, [''], ["genes.tsv.gz", "genes.tsv", "features.tsv.gz", "features.tsv"])
    else:
        p1, p2, p3 = fname.partition("matrix")
        if p2 == '' and p3 == '':
            barcode_file = _enumerate_files(path, [p1, ''], [".barcodes.tsv.gz", ".barcodes.tsv", ".cells.tsv.gz", ".cells.tsv", "_barcode.tsv", ".barcodes.txt"])
            feature_file = _enumerate_files(path, [p1, ''], [".genes.tsv.gz", ".genes.tsv", ".features.tsv.gz", ".features.tsv", "_gene.tsv", ".genes.txt"])
        else:
            barcode_file = _enumerate_files(path, [p1, '', p3, ''], ["barcodes", "cells"], [".tsv.gz", ".tsv"])
            feature_file = _enumerate_files(path, [p1, '', p3, ''], ["genes", "features"], [".tsv.gz", ".tsv"])

    if barcode_file is None:
        raise ValueError("Cannot find barcode file!")
    if feature_file is None:
        raise ValueError("Cannot find feature file!")

    return barcode_file, feature_file


def _load_barcode_metadata(barcode_file: str, sep: str = "\t") -> Tuple[pd.DataFrame, str]:
    """ Load cell barcode information """
    format_type = None
    barcode_metadata = pd.read_csv(barcode_file, sep=sep, header=None)

    if "cellkey" in barcode_metadata.iloc[0].values:
        # HCA DCP format
        barcode_metadata = pd.DataFrame(data = barcode_metadata.iloc[1:].values, columns = barcode_metadata.iloc[0].values)
        barcode_metadata.rename(columns={"cellkey": "barcodekey"}, inplace=True)
        format_type = "HCA DCP"
    elif "barcodekey" in barcode_metadata.iloc[0].values:
        # Pegasus format
        barcode_metadata = pd.DataFrame(data = barcode_metadata.iloc[1:].values, columns = barcode_metadata.iloc[0].values)
        format_type = "Pegasus"
    else:
        # Other format, only one column containing the barcode is expected
        barcode_metadata = pd.read_csv(barcode_file, sep="\t", header=None, names=["barcodekey"])
        format_type = "other"

    return barcode_metadata, format_type


def _load_feature_metadata(feature_file: str, format_type: str, sep: str = "\t") -> Tuple[pd.DataFrame, str]:
    """ Load feature information """
    feature_metadata = pd.read_csv(feature_file, sep=sep, header=None)

    if format_type == "HCA DCP":
        # HCA DCP format genes.tsv.gz
        series = feature_metadata.iloc[0]
        assert "featurekey" in series.values and "featurename" in series.values
        series[series.eq("featurekey")] = "featureid"
        series[series.eq("featurename")] = "featurekey"
        feature_metadata = pd.DataFrame(data = feature_metadata.iloc[1:].values, columns = series.values)
    elif format_type == "Pegasus":
        # Pegasus format features.tsv.gz
        series = feature_metadata.iloc[0]
        assert "featurekey" in series.values
        feature_metadata = pd.DataFrame(data = feature_metadata.iloc[1:].values, columns = series.values)
    else:
        # Other format
        assert format_type == "other" and feature_metadata.shape[1] <= 3
        # shape[1] = 3 -> 10x v3, 2 -> 10x v2, 1 -> scumi, dropEst, BUStools
        if feature_metadata.shape[1] > 1:
            col_names = ["featureid", "featurekey", "featuretype"]
            format_type = "10x v3" if feature_metadata.shape[1] == 3 else "10x v2"
            feature_metadata.columns = col_names[:feature_metadata.shape[1]]
        else:
            assert feature_metadata.shape[1] == 1
            feature_metadata.columns = ["featurekey"]
            # Test if in scumi format
            if feature_metadata.iloc[0, 0].find('_') >= 0:
                format_type = "scumi"
                values = feature_metadata.values[:, 0].astype(str)
                arr = np.array(np.char.split(values, sep="_", maxsplit=1).tolist())
                feature_metadata = pd.DataFrame(data={"featureid": arr[:, 0], "featurekey": arr[:, 1]})
            else:
                format_type = "dropEst or BUStools"

    return feature_metadata, format_type


def load_one_mtx_file(path: str, file_name: str, genome: str, modality: str) -> UnimodalData:
    """Load one gene-count matrix in mtx format into a UnimodalData object
    """
    try:
        from pegasusio.cylib.io import read_mtx
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.io'")

    fname = re.sub('(.mtx|.mtx.gz)$', '', file_name)
    barcode_file, feature_file = _locate_barcode_and_feature_files(path, fname)

    barcode_metadata, format_type = _load_barcode_metadata(barcode_file)
    feature_metadata, format_type = _load_feature_metadata(feature_file, format_type)
    logger.info(f"Detected mtx file in {format_type} format.")

    mtx_file = os.path.join(path, file_name)
    if file_name.endswith(".gz"):
        mtx_fifo = os.path.join(tempfile.gettempdir(), file_name + ".fifo")
        if os.path.exists(mtx_fifo):
            os.unlink(mtx_fifo)
        os.mkfifo(mtx_fifo)
        subprocess.Popen(f"gunzip -c {shlex.quote(mtx_file)} > {shlex.quote(mtx_fifo)}", shell = True)
        row_ind, col_ind, data, shape = read_mtx(mtx_fifo)
        os.unlink(mtx_fifo)
    else:
        row_ind, col_ind, data, shape = read_mtx(mtx_file)

    if shape[1] == barcode_metadata.shape[0]: # Column is barcode, swap the coordinates
        row_ind, col_ind = col_ind, row_ind
        shape = (shape[1], shape[0])

    mat = csr_matrix((data, (row_ind, col_ind)), shape = shape)
    mat.eliminate_zeros()

    unidata = UnimodalData(barcode_metadata, feature_metadata, {"X": mat}, {"genome": genome, "modality": modality})
    if format_type == "10x v3" or format_type == "10x v2":
        unidata.separate_channels()

    return unidata


def _locate_mtx_file(path: str) -> str:
    """ Locate one mtx file in the path directory; first choose matrix.mtx.gz or matrix.mtx if multiple mtx files exist."""
    file_names = []
    with os.scandir(path) as dir_iter:
        for dir_entry in dir_iter:
            file_name = dir_entry.name
            if file_name.endswith(".mtx.gz") or file_name.endswith(".mtx"):
                if file_name == "matrix.mtx.gz" or file_name == "matrix.mtx":
                    return file_name
                file_names.append(file_name)
    return file_names[0] if len(file_names) > 0 else None


def _parse_dir_name(dirname: str, default_modality: str) -> Tuple[str, str]:
    """ Parse directory name to see if we can separate genome and modality """
    genome, sep, modality = dirname.rpartition('-')
    if genome == "":
        genome = dirname
        modality = default_modality
    return genome, modality


def load_mtx_file(path: str, genome: str = None, modality: str = None) -> MultimodalData:
    """Load gene-count matrix from Market Matrix files (10x v2, v3 and HCA DCP formats)

    Parameters
    ----------

    path : `str`
        Path to mtx files. The directory implied by path should either contain matrix, feature and barcode information, or folders containing these information.
    genome : `str`, optional (default: None)
        Genome name of the matrix. If None, genome will be inferred from path.
    modality: `str`, optional (default: None)
        Modality, choosing from 'rna', 'citeseq', 'hashing', 'tcr', 'bcr', 'crispr' or 'atac'. If None, use 'rna' as default.

    Returns
    -------

    A MultimodalData object containing (genome, UnimodalData) pairs.

    Examples
    --------
    >>> io.load_mtx_file('example.mtx.gz', genome = 'mm10')
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"{path} does not exist!")

    orig_file = path
    if os.path.isdir(orig_file):
        path = orig_file.rstrip('/')
        file_name = _locate_mtx_file(path)
    else:
        if (not orig_file.endswith(".mtx")) and (not orig_file.endswith(".mtx.gz")):
            raise ValueError(f"File {orig_file} does not end with suffix .mtx or .mtx.gz!")
        path, file_name = os.path.split(orig_file)

    data = MultimodalData()

    if modality is None:
        modality = "rna"

    if file_name is not None:
        if genome is None:
            genome = "unknown"
        data.add_data(
            load_one_mtx_file(
                path,
                file_name,
                genome,
                modality
            ),
        )
    else:
        for dir_entry in os.scandir(path):
            if dir_entry.is_dir():
                file_name = _locate_mtx_file(dir_entry.path)
                if file_name is None:
                    raise ValueError(f"Folder {dir_entry.path} does not contain a mtx file!")
                dgenome, dmodality = _parse_dir_name(dir_entry.name, modality)
                data.add_data(load_one_mtx_file(dir_entry.path, file_name, dgenome, dmodality))

    return data



def _write_mtx(unidata: UnimodalData, output_dir: str, precision: int):
    """ Write Unimodal data to mtx
    """
    try:
        from pegasusio.cylib.io import write_mtx
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.io'")

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    for key in unidata.list_keys():
        matrix = unidata.matrices[key]
        mtx_file = os.path.join(output_dir, ("matrix" if key == "X" else key) + ".mtx.gz")
        fifo_file = mtx_file + ".fifo"
        if os.path.exists(fifo_file):
            os.unlink(fifo_file)
        os.mkfifo(fifo_file)
        pobj = subprocess.Popen(f"gzip < {shlex.quote(fifo_file)} > {shlex.quote(mtx_file)}", shell = True)
        write_mtx(fifo_file, matrix.data, matrix.indices, matrix.indptr, matrix.shape[0], matrix.shape[1], precision = precision) # matrix is cell x gene csr_matrix, will write as gene x cell
        assert pobj.wait() == 0
        os.unlink(fifo_file)
        logger.info(f"{mtx_file} is written.")

    unidata.barcode_metadata.to_csv(os.path.join(output_dir, "barcodes.tsv.gz"), sep = '\t')
    logger.info("barcodes.tsv.gz is written.")

    unidata.feature_metadata.to_csv(os.path.join(output_dir, "features.tsv.gz"), sep = '\t')
    logger.info("features.tsv.gz is written.")

    logger.info(f"Mtx for {unidata.get_uid()} is written.")


def write_mtx_file(data: MultimodalData, output_directory: str, precision: int = 2):
    """ Write output to mtx files in output_directory
    """
    output_dir = os.path.abspath(output_directory)
    os.makedirs(output_dir, exist_ok = True)

    for key in data.list_data():
        _write_mtx(data.get_data(key), os.path.join(output_dir, key), precision)

    logger.info("Mtx files are written.")



def load_csv_file(
    input_csv: str,
    sep: str = ",",
    genome: str = None,
    modality: str = None,
) -> MultimodalData:
    """Load count matrix from a CSV-style file, such as CSV file or DGE style tsv file.

    Parameters
    ----------

    input_csv : `str`
        The CSV file, gzipped or not, containing the count matrix.
    sep: `str`, optional (default: ',')
        Separator between fields, either ',' or '\t'.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead.
    modality: `str`, optional (default None)
        Modality. If None, use "rna" instead.

    Returns
    -------

    A MultimodalData object containing a (genome, UnimodalData) pair.

    Examples
    --------
    >>> io.load_csv_file('example_ADT.csv')
    >>> io.load_csv_file('example.umi.dge.txt.gz', genome = 'GRCh38', sep = '\t')
    """
    try:
        from pegasusio.cylib.io import read_csv
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.io'")

    if not os.path.exists(input_csv):
        raise FileNotFoundError(f"File {input_csv} does not exist!")

    barcode_metadata = feature_metadata = None

    input_csv = os.path.abspath(input_csv)
    path = os.path.dirname(input_csv)
    fname = os.path.basename(input_csv)

    barcode_file = os.path.join(path, "cells.csv")
    if not os.path.isfile(barcode_file):
        barcode_file += ".gz"
    feature_file = os.path.join(path, "genes.csv")
    if not os.path.isfile(feature_file):
        feature_file += ".gz"

    if os.path.isfile(barcode_file) and os.path.isfile(feature_file):
        barcode_metadata, format_type = _load_barcode_metadata(barcode_file, sep = sep)
        feature_metadata, format_type = _load_feature_metadata(feature_file, format_type, sep = sep)
        assert format_type == "HCA DCP"

    if input_csv.endswith(".gz"):
        csv_fifo = os.path.join(tempfile.gettempdir(), fname + ".fifo")
        if os.path.exists(csv_fifo):
            os.unlink(csv_fifo)
        os.mkfifo(csv_fifo)
        subprocess.Popen(f"gunzip -c {shlex.quote(input_csv)} > {shlex.quote(csv_fifo)}", shell = True)
        row_ind, col_ind, data, shape, rowkey, rownames, colnames = read_csv(csv_fifo, sep)
        os.unlink(csv_fifo)
    else:
        row_ind, col_ind, data, shape, rowkey, rownames, colnames = read_csv(input_csv, sep)

    if rowkey == "cellkey":
        # HCA format
        assert (barcode_metadata is not None) and (feature_metadata is not None) and (barcode_metadata.shape[0] == shape[0]) and (feature_metadata.shape[0] == shape[1]) and \
               ((barcode_metadata["barcodekey"].values != np.array(rownames)).sum() == 0) and ((feature_metadata["featureid"].values != np.array(colnames)).sum() == 0)
        mat = csr_matrix((data, (row_ind, col_ind)), shape = shape)
    else:
        mat = csr_matrix((data, (col_ind, row_ind)), shape = (shape[1], shape[0]))
        if barcode_metadata is None:
            barcode_metadata = {"barcodekey": colnames}
        else:
            assert (barcode_metadata.shape[0] == shape[1]) and ((barcode_metadata["barcodekey"].values != np.array(colnames)).sum() == 0)
        if feature_metadata is None:
            feature_metadata = {"featurekey": rownames}
        else:
            assert (feature_metadata.shape[0] == shape[0]) and ((feature_metadata["featurekey"].values != np.array(rownames)).sum() == 0)

    genome = genome if genome is not None else "unknown"
    modality = modality if modality is not None else "rna"

    if modality == "citeseq":
        unidata = CITESeqData(barcode_metadata, feature_metadata, {"raw.count": mat}, {"genome": genome, "modality": modality})
    else:
        unidata = UnimodalData(barcode_metadata, feature_metadata, {"X": mat}, {"genome": genome, "modality": modality})

    data = MultimodalData(unidata)

    return data


def _write_scp_metadata(unidata: UnimodalData, output_name: str, precision: int = 2) -> None:
    """ Write metadata for SCP
    """
    metadata_list = []
    datatype_list = []
    for col_name in unidata.obs.columns:
        metadata_list.append("\t" + col_name)
        if unidata.obs[col_name].dtype.kind in ['i', 'u', 'f', 'c']:
            datatype_list.append("\tnumeric")
        else:
            datatype_list.append("\tgroup")

    metadata_file = f"{output_name}.scp.metadata.txt"
    with open(metadata_file, "w") as fout:
        fout.write(f"NAME{''.join(metadata_list)}\n")
        fout.write(f"TYPE{''.join(datatype_list)}\n")

    unidata.obs.to_csv(metadata_file, sep="\t", na_rep = "", float_format = f"%.{precision}f", header=False, mode="a")
    logger.info(f"Metadata file {metadata_file} is written.")


def _write_scp_coords(unidata: UnimodalData, output_name: str, precision: int = 2) -> None:
    """ Write coordinates for each visualization
    """
    float_fmt = f"%.{precision}f"
    for basis in unidata.obsm.keys():
        basis_X = unidata.obsm[basis]
        if not basis.startswith("X_") or basis_X.ndim != 2 or basis_X.shape[1] < 2:
            continue
        coords = ["X", "Y"] if basis_X.shape[1] == 2 else ["X", "Y", "Z"]
        coord_file = f"{output_name}.scp.{basis}.coords.txt"
        with open(coord_file, "w") as fout:
            fout.write(f"NAME\t{chr(9).join(coords)}\n")
            fout.write(f"TYPE\t{chr(9).join(['numeric'] * len(coords))}\n")
        df_out = pd.DataFrame(
            basis_X[:, 0: len(coords)],
            columns=coords,
            index=unidata.obs_names,
        )
        df_out.to_csv(coord_file, sep="\t", na_rep = "", float_format = float_fmt, header=False, mode="a")
        logger.info(f"Coordinate file {coord_file} is written.")


def _write_scp_expression(unidata: UnimodalData, output_name: str, is_sparse: bool, precision: int = 2) -> None:
    """ Only write the main matrix X
    """
    try:
        from pegasusio.cylib.io import write_mtx, write_dense
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.io'")

    matrix = unidata.get_matrix("X")
    if is_sparse:
        barcode_file = f"{output_name}.scp.barcodes.tsv"
        with open(barcode_file, "w") as fout:
            fout.write("\n".join(unidata.obs_names) + "\n")
        logger.info(f"Barcode file {barcode_file} is written.")

        feature_file = f"{output_name}.scp.features.tsv"

        gene_names = unidata.var_names.values
        gene_ids = unidata.var["featureid"].values if "featureid" in unidata.var else (unidata.var["gene_ids"] if "gene_ids" in unidata.var else gene_names)

        df = pd.DataFrame({"gene_names": gene_names, "gene_ids": gene_ids})[["gene_ids", "gene_names"]]
        df.to_csv(feature_file, sep="\t", header=False, index=False)
        logger.info(f"Feature file {feature_file} is written.")

        mtx_file = f"{output_name}.scp.matrix.mtx"
        write_mtx(mtx_file, matrix.data, matrix.indices, matrix.indptr, matrix.shape[0], matrix.shape[1], precision = precision) # matrix is cell x gene csr_matrix, will write as gene x cell
        logger.info(f"Matrix file {mtx_file} is written.")
    else:
        expr_file = f"{output_name}.scp.expr.txt"
        matrix = matrix.T.tocsr() # convert to gene x cell
        write_dense(expr_file, unidata.obs_names.values, unidata.var_names.values, matrix.data, matrix.indices, matrix.indptr, matrix.shape[0], matrix.shape[1], precision = precision)
        logger.info(f"Dense expression file {expr_file} is written.")


def write_scp_file(data: MultimodalData, output_name: str, is_sparse: bool = True, precision: int = 2) -> None:
    """Generate outputs for single cell portal from a MultimodalData object

    Parameters
    ----------
    data: MultimodalData
        A MultimodalData object.

    output_name: ``str``
        Name prefix for output files.

    is_sparse: ``bool``, optional, default: ``True``
        If ``True``, enforce the count matrix to be sparse after written into files.

    precision: ``int``, optional, default: ``2``
        Round numbers to ``precision`` decimal places.

    Returns
    -------
    ``None``

    If data contains only one modality, files are generated as follows:
        * ``output_name.scp.basis.coords.txt``, where ``basis`` is for each key in ``adata.obsm`` field.
        * ``output_name.scp.metadata.txt``.
        * Gene expression files:
            * If in sparse format:
                * ``output_name.scp.features.tsv``, information on genes;
                * ``output_name.scp.barcodes.tsv``, information on cell barcodes;
                * ``output_name.scp.matrix.mtx``, count matrix.
            * If not in sparse:
                * ``output_name.scp.expr.txt``.

    Otherwise, under the directory os.path.dirname(output_name), we will create separate folders per key and each separate folder has its own files in the format above.

    Examples
    --------
    >>> io.write_scp_file(data, output_name = "scp_result")
    """
    output_name = os.path.abspath(output_name)

    valid_keys = []
    for key in data.list_data():
        try:
            mat = data.get_data(key).get_matrix("X")
            if mat.data.size > 0:
                valid_keys.append(key)
        except ValueError:
            pass

    if len(valid_keys) == 1:
        unidata = data.get_data(valid_keys[0])
        _write_scp_metadata(unidata, output_name, precision = precision)
        _write_scp_coords(unidata, output_name, precision = precision)
        _write_scp_expression(unidata, output_name, is_sparse, precision = precision)
    else:
        path = os.path.dirname(output_name)
        fname = os.path.basename(output_name)
        for key in valid_keys:
            subpath = os.path.join(path, key)
            if not os.path.isdir(subpath):
                os.path.mkdir(subpath)
            out_new_name = os.path.join(subpath, fname)

            unidata = data.get_data(key)
            _write_scp_metadata(unidata, out_new_name, precision = precision)
            _write_scp_coords(unidata, out_new_name, precision = precision)
            _write_scp_expression(unidata, out_new_name, is_sparse, precision = precision)

    logger.info("write_scp_file is done.")



# cython: language_level=3, c_string_type=str, c_string_encoding=default

import numpy as np

from libc.stdlib cimport malloc, free, atoi
from libc.stdio cimport fopen, fclose, getline, FILE, fscanf, sscanf, fprintf, fread, fseek, SEEK_CUR, SEEK_SET#, printf
from libc.string cimport strncmp, strlen, strtok, strcmp, memcpy, memset
from libc.math cimport pow

cimport cython

ctypedef unsigned char uchar

ctypedef fused data_type:
    int
    long
    float
    double

ctypedef fused indices_type:
    int
    long

ctypedef fused indptr_type:
    int
    long



cdef const char* header_real = b"%%MatrixMarket matrix coordinate real general"
cdef const char* header_int = b"%%MatrixMarket matrix coordinate integer general"
cdef const char* metadata = b"%metadata_json: {\"software\": \"pegasusio\"}"


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple read_mtx(char* mtx_file):
    cdef FILE* fi = fopen(mtx_file, "r")
    cdef char* line = NULL
    cdef size_t size
    cdef char isreal = 0

    assert getline(&line, &size, fi) >= 0
    if strncmp(line, header_real, strlen(header_real)) == 0:
        isreal = 1
    elif strncmp(line, header_int, strlen(header_int)) == 0:
        isreal = 0
    else:
        raise ValueError("Cannot recognize header line : " + line)

    while getline(&line, &size, fi) >=0:
        if line[0] != b'%':
            break

    cdef size_t M, N, L, i

    assert line[0] != b'%'
    assert sscanf(line, "%zu %zu %zu", &M, &N, &L) >= 0
    free(line)

    row_ind = np.zeros(L, dtype = np.intc)
    col_ind = np.zeros(L, dtype = np.intc)
    data = np.zeros(L, dtype = np.float32 if isreal else np.intc)

    cdef int[:] rview = row_ind, cview = col_ind
    cdef int[:] dview_int
    cdef float[:] dview_float

    cdef int x, y, value_int
    cdef float value_float

    if isreal:
        dview_float = data
        for i in range(L):
            assert fscanf(fi, "%d %d %f", &x, &y, &value_float) >= 0
            rview[i] = x - 1
            cview[i] = y - 1
            dview_float[i] = value_float
    else:
        dview_int = data
        for i in range(L):
            assert fscanf(fi, "%d %d %d", &x, &y, &value_int) >= 0
            rview[i] = x - 1
            cview[i] = y - 1
            dview_int[i] = value_int

    fclose(fi)

    return row_ind, col_ind, data, (M, N)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void write_mtx(char* mtx_file, data_type[:] data, indices_type[:] indices, indptr_type[:] indptr, int M, int N, int precision = 2):
    """ Input is csr_matrix internal representation, cell by gene; Output will be gene by cell
    """
    cdef FILE* fo = fopen(mtx_file, "w")
    cdef str fmt_str = ""

    if (data_type is float) or (data_type is double):
        fprintf(fo, "%s\n", header_real)
        fmt_str = f"%d %d %.{precision}f\n"
    else:
        fprintf(fo, "%s\n", header_int)
        fmt_str = "%d %d %d\n"

    cdef const char* fmt = fmt_str
    cdef Py_ssize_t i, j

    fprintf(fo, "%s\n", metadata)
    fprintf(fo, "%d %d %zd\n", N, M, <long>data.size)

    for i in range(M):
        for j in range(indptr[i], indptr[i + 1]):
            fprintf(fo, fmt, indices[j] + 1, i + 1, data[j])

    fclose(fo)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef int is_zero(char* pch):
    cdef int i = 0
    while pch[i] != 0:
        if pch[i] >= b'1' and pch[i] <= b'9':
            return 0
        i += 1
    return 1

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple read_csv(char* csv_file, char* delimiters):
    cdef FILE* fi = fopen(csv_file, "r")
    cdef char* line = NULL
    cdef char* pch = NULL
    cdef size_t size

    cdef str row_key
    cdef list colnames = [], rownames = [], row_ind = [], col_ind = [], data_list = []
    cdef int M = 0, N = 0
    cdef size_t L = 0
    cdef Py_ssize_t i

    assert getline(&line, &size, fi) >= 0
    pch = strtok(line, delimiters)
    assert pch != NULL
    row_key = pch

    pch = strtok(NULL, delimiters)
    while pch != NULL:
        colnames.append(pch)
        N += 1
        pch = strtok(NULL, delimiters)

    if N == 0:
        raise ValueError(f"File {csv_file} contains no columns!")

    colnames[N - 1] = colnames[N - 1].rstrip("\n\r")

    while getline(&line, &size, fi) >=0:
        if line[0] < 32 or line[0] == 127:
            continue
        pch = strtok(line, delimiters)
        assert pch != NULL
        rownames.append(pch)
        for i in range(N):
            pch = strtok(NULL, delimiters)
            assert pch != NULL
            if is_zero(pch) == 0:
                row_ind.append(M)
                col_ind.append(i)
                data_list.append(pch)
                if i == N - 1:
                    data_list[L] = data_list[L].rstrip("\r\n")
                L += 1
        M += 1

    free(line)
    fclose(fi)

    try:
        data = np.array(data_list, dtype = np.intc)
    except ValueError:
        data = np.array(data_list, dtype = np.float32)

    return row_ind, col_ind, data, (M, N), row_key, rownames, colnames


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef void write_dense(char* output_file, str[:] barcodes, str[:] features, data_type[:] data, indices_type[:] indices, indptr_type[:] indptr, int M, int N, int precision = 2):
    """ Input must be csr_matrix internal representation, gene by cell (X.T.tocsr()); Output will be gene by cell
    """
    cdef FILE* fo = fopen(output_file, "w")
    cdef str fmt_str = f"\t%.{precision}f" if (data_type is float) or (data_type is double) else "\t%d"

    cdef const char* fmt = fmt_str
    cdef Py_ssize_t i, j, k, fr

    fprintf(fo, "GENE");
    for i in range(N):
        fprintf(fo, "\t%s", <char*>barcodes[i])
    fprintf(fo, "\n")

    for i in range(M):
        fprintf(fo, "%s", <char*>features[i]);
        fr = 0
        for j in range(indptr[i], indptr[i + 1]):
            for k in range(fr, indices[j]):
                fprintf(fo, "\t0")
            fprintf(fo, fmt, data[j])
            fr = indices[j] + 1
        for k in range(fr, N):
            fprintf(fo, "\t0")
        fprintf(fo, "\n")

    fclose(fo)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef swapbytes(uchar* bytes_buffer, int s):
    cdef int i, j
    cdef uchar tmp

    for i in range(s >> 1):
        j = s - i - 1
        tmp = bytes_buffer[i]
        bytes_buffer[i] = bytes_buffer[j]
        bytes_buffer[j] = tmp


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef tuple read_fcs(char* fcs_file):
    """ This function is inspired by the python package fcsparser (https://github.com/eyurtsev/fcsparser)
    """
    cdef FILE* fi = fopen(fcs_file, "rb")

    # parse HEADER segment
    cdef char[7] fcs_format
    cdef char[9] header_field
    cdef size_t text_start, text_end, data_start, data_end, analysis_start, analysis_end

    fcs_format[6] = header_field[8] = 0
    assert fread(<void*>fcs_format, 1, 6, fi) == 6
    if strcmp(fcs_format, "FCS2.0") and strcmp(fcs_format, "FCS3.0") and strcmp(fcs_format, "FCS3.1"):
        fclose(fi)
        raise ValueError(f"Detected unsupported FCS format '{fcs_format}'!")

    assert fseek(fi, 4, SEEK_CUR) == 0
    assert fread(<void*>header_field, 1, 8, fi) == 8
    text_start = atoi(header_field)
    assert fread(<void*>header_field, 1, 8, fi) == 8
    text_end = atoi(header_field)
    assert fread(<void*>header_field, 1, 8, fi) == 8
    data_start = atoi(header_field)
    assert fread(<void*>header_field, 1, 8, fi) == 8
    data_end = atoi(header_field)
    assert fread(<void*>header_field, 1, 8, fi) == 8
    analysis_start = atoi(header_field)
    assert fread(<void*>header_field, 1, 8, fi) == 8
    analysis_end = atoi(header_field)

    if text_start < 58 or text_start >= text_end:
        fclose(fi)
        raise ValueError(f"Detected invalid TEXT segment start and end offsets: [{text_start}, {text_end}]!")

    if text_end == data_start:
        text_end -= 1

    # parse TEXT segment
    cdef char delim
    cdef size_t text_bytes, pos
    cdef uchar* text_segment

    unicode_source = np.zeros(100000, dtype = np.uint8)

    cdef uchar[:] sview = unicode_source
    cdef int s_pos # unicode source position
    cdef dict metadata = {}
    cdef bytes value_bytes
    cdef str key, value

    text_bytes = text_end - text_start + 1
    text_segment = <uchar*>malloc(text_bytes)
    assert text_segment != NULL
    assert fseek(fi, text_start, SEEK_SET) == 0
    assert fread(<void*>text_segment, 1, text_bytes, fi) == text_bytes

    # determine the end position of text segment
    delim = text_segment[0]
    pos = text_bytes - 1
    while text_segment[pos] != delim:
        pos -= 1
    text_bytes = pos + 1

    # generate key:value dictionary
    pos = 1
    s_pos = 0
    key = ""
    while pos < text_bytes:
        if text_segment[pos] == delim:
            if pos + 2 < text_bytes and text_segment[pos + 1] == delim: # delimiter escaping cannot happen at the end of the text segment
                sview[s_pos] = text_segment[pos]
                s_pos += 1
                pos += 1
            else:
                if s_pos == 0:
                    free(text_segment)
                    fclose(fi)
                    raise ValueError("Keywords and keyword values must have lengths greater than zero!")

                value_bytes = bytes(unicode_source[0:s_pos])
                try:
                    value = value_bytes.decode(encoding = 'utf-8')
                except UnicodeDecodeError:
                    value = value_bytes.decode(encoding = 'latin-1')

                s_pos = 0
                if key == "":
                    key = value.upper() # keywords are case insensitive; convert them to upper case
                else:
                    metadata[key] = value
                    key = ""
        else:
            sview[s_pos] = text_segment[pos]
            s_pos += 1

        pos += 1

        if s_pos >= 100000:
            free(text_segment)
            fclose(fi)
            raise NotImplementedError("Keyword or keyward value length cannot exceed 10,000!")

    free(text_segment)

    if key != "":
        fclose(fi)
        raise ValueError(f"Keyword {key} does not have a keyword value!")

    # sanity checks
    cdef size_t _start, _end

    _start = _end = 0
    if "$BEGINDATA" in metadata:
        _start = int(metadata.pop("$BEGINDATA"))
    if "$ENDDATA" in metadata:
        _end = int(metadata.pop("$ENDDATA"))
    if data_start == 0 and data_end == 0:
        data_start = _start
        data_end = _end
    if data_start == 0:
        print("Warning: Could not detect a DATA segment!")

    _start = _end = 0
    if "$BEGINANALYSIS" in metadata:
        _start = int(metadata.pop("$BEGINANALYSIS"))
    if "$ENDANALYSIS" in metadata:
        _end = int(metadata.pop("$ENDANALYSIS"))
    if analysis_start == 0 and analysis_end == 0:
        analysis_start = _start
        analysis_end = _end
    if analysis_start > 0:
        print("Warning: ANALYSIS segment would be omitted since current implementation does not support parsing of this segment.")

    if "$BEGINSTEXT" in metadata:
        if metadata.pop("$BEGINSTEXT") != "0":
            print("Warning: Detected a supplemental TEXT segment. This segment will be ignored under current implementation.")
    if "$ENDSTEXT" in metadata:
        del metadata["$ENDSTEXT"]

    # parse data segment
    cdef int M, N # M parameters and N events; matrix is N x M
    cdef str endian
    cdef char datatype
    cdef dict feature_metadata
    cdef Py_ssize_t i, j, k
    cdef str keyword, tmpstr

    for keyword in ["$NEXTDATA", "$MODE", "$PAR", "$TOT", "$BYTEORD", "$DATATYPE"]:
        if keyword not in metadata:
            fclose(fi)
            raise ValueError(f"Cannot detect requried keyword {keyword} in TEXT segment!")

    if metadata.pop("$NEXTDATA") != "0":
        print("Warning: Detected more than one data sets. Current implementation will only parse the first data set.")

    if metadata.pop("$MODE") != "L":
        fclose(fi)
        raise NotImplementedError("Current implementation can only handle $MODE = L!")

    endian = "little"
    if metadata.pop("$BYTEORD") in ["4,3,2,1", "2,1"]:
        endian = "big"

    datatype = ord(metadata.pop("$DATATYPE")[0])
    if datatype != b'I' and datatype != b'F' and datatype != b'D' and datatype != b'A':
        fclose(fi)
        raise ValueError(f"$DATATYPE {datatype} does not in 'IFDA'!")


    M = int(metadata.pop("$PAR"))
    N = int(metadata.pop("$TOT"))

    feature_metadata = {"featurekey": np.empty(M, dtype = object), "featureid": np.empty(M, dtype = object)}

    cdef str[:] fkview = feature_metadata["featurekey"]
    cdef str[:] fiview = feature_metadata["featureid"]

    PBs = np.zeros(M, dtype = np.intc)
    PRs = np.zeros(M, dtype = np.intc)
    PEs = np.empty(M, dtype = object)
    log_decades = np.zeros(M, dtype = np.float32)
    linear_values = np.zeros(M, dtype = np.float32)
    bit_masks = np.zeros(M, dtype = np.intc)

    cdef int[:] PBsview = PBs
    cdef int[:] PRsview = PRs
    cdef str[:] PEsview = PEs
    cdef float[:] ldview = log_decades
    cdef float[:] lvview = linear_values
    cdef int[:] bmview = bit_masks

    cdef str ldstr, lvstr
    cdef int total_bytes_per_row = 0

    for i in range(1, M + 1):
        for tmpstr in ["B", "E", "N", "R"]:
            keyword = f"$P{i}{tmpstr}"
            if keyword not in metadata:
                fclose(fi)
                raise ValueError(f"Cannot detect requried keyword {keyword} in TEXT segment!")

        pos = i - 1
        fiview[pos] = metadata.pop(f"$P{i}N")
        fkview[pos] = metadata.pop(f"$P{i}S", fiview[pos])
        PEsview[pos] = metadata.pop(f"$P{i}E")

        tmpstr = metadata.pop(f"$P{i}R")
        try:
            PRsview[pos] = int(tmpstr)
        except ValueError:
            PRsview[pos] = int(float(tmpstr))
            print(f"Warning: $P{i}R should be an integer! Rounded {tmpstr} to {PRsview[pos]}.")

        ldstr, lvstr = PEsview[pos].split(",")
        ldview[pos] = float(ldstr)
        lvview[pos] = float(lvstr)
        assert ldview[pos] >= 0.0 and lvview[pos] >= 0.0

        keyword = f"$P{i}B"

        if metadata[keyword] == "*":
            fclose(fi)
            if datatype == b'A':
                raise NotImplementedError(f"Current implementation does not support {keyword} = *!")
            else:
                raise ValueError(f"$DATATYPE '{datatype}' should have {keyword} = *!")

        PBsview[pos] = int(metadata.pop(keyword))

        if datatype == b'F':
            if PBsview[pos] != 32:
                fclose(fi)
                raise ValueError(f"$DATATYPE '{datatype}' must have {keyword} = 32 ({keyword} = {PBsview[pos]})!")
            PBsview[pos] = PBsview[pos] // 8
            if ldview[pos] > 0.0 or lvview[pos] > 0.0:
                fclose(fi)
                raise ValueError(f"$DATATYPE '{datatype}' must have $P{i}E = 0,0 ($P{i}E = {PEsview[pos]})!")
        elif datatype == b'D':
            if PBsview[pos] != 64:
                fclose(fi)
                raise ValueError(f"$DATATYPE '{datatype}' must have {keyword} = 64 ({keyword} = {PBsview[pos]})!")
            PBsview[pos] = PBsview[pos] // 8
            if ldview[pos] > 0.0 or lvview[pos] > 0.0:
                fclose(fi)
                raise ValueError(f"$DATATYPE '{datatype}' must have $P{i}E = 0,0 ($P{i}E = {PEsview[pos]})!")
        else:
            if datatype == b'I':
                if PBsview[pos] % 8 != 0:
                    fclose(fi)
                    raise NotImplementedError(f"Current implementation requires that {keyword} must be a multiple of 8 ({keyword} = {PBsview[pos]}) for $DATATYPE '{datatype}'!")
                PBsview[pos] = PBsview[pos] // 8
                if PBsview[pos] > 4:
                    fclose(fi)
                    raise NotImplementedError(f"Current implementation requires that {keyword} be no more than 4 bytes for $DATATYPE '{datatype}'!")
                bmview[pos] = 2 ** (PRsview[pos] - 1).bit_length() - 1
            else:
                # datatype must be 'A'
                if PBsview[pos] > 9:
                    fclose(fi)
                    raise NotImplementedError(f"Current implementation does not support {keyword} > 9 for $DATATYPE '{datatype}'!")

            if ldview[pos] > 0.0 or lvview[pos] > 0.0:
                if ldview[pos] == 0.0:
                    fclose(fi)
                    raise ValueError(f"We require f1 > 0 for $PnE/f1,f2/({PEsview[pos]})!")
                if lvview[pos] == 0.0:
                    lvview[pos] = 1.0
                    print(f"Warning: detected f2 = 0 for $P{i}E/{PEsview[pos]}/, which is not correct. Set f2 = 1 instead.")

        total_bytes_per_row += PBsview[pos]


    # SEEK data start
    data = np.zeros((N, M), dtype = np.float32)
    cdef float[:, :] data_view = data

    cdef size_t buffer_size = (<size_t>N) * (<size_t>total_bytes_per_row)
    cdef size_t data_size = data_end - data_start + 1
    if buffer_size != data_size and buffer_size != data_size - 1: # buffer_size == data_size - 1 because [data_start, data_end)
        print(f"Warning: Data size recorded in the DATA segment {data_end - data_start + 1} does not match the actual data size {buffer_size}.")

    cdef uchar* bytes_buffer = <uchar*>malloc(buffer_size + 1)
    assert bytes_buffer != NULL

    cdef uchar* buf = bytes_buffer

    from sys import byteorder
    cdef bint should_swap = endian != byteorder

    cdef int value_int
    cdef double value_double

    assert fseek(fi, data_start, SEEK_SET) == 0
    assert fread(<void*>bytes_buffer, 1, buffer_size, fi) == buffer_size

    for i in range(N):
        for j in range(M):
            if datatype != b'A' and should_swap:
                swapbytes(buf, PBsview[j])

            if datatype == b'D':
                memcpy(<void*>(&value_double), <void*>buf, PBsview[j])
                data_view[i, j] = <float>value_double
            elif datatype == b'F':
                memcpy(<void*>(&data_view[i, j]), <void*>buf, PBsview[j])
            else:
                if datatype == b'I':
                    memcpy(<void*>(&value_int), <void*>buf, PBsview[j])
                    value_int = value_int & bmview[j] # masking undefined bits
                else:
                    # datatype must be 'A'
                    value_int = 0
                    for k in range(PBsview[j]):
                        value_int = value_int * 10 + ((<char>buf[k]) - b'0')

                if ldview[j] == 0.0 and lvview[j] == 0.0:
                    data_view[i, j] = <float>value_int
                else:
                    data_view[i, j] = (<float>pow(10.0, ldview[j] * value_int / PRsview[j])) * lvview[j]

            buf += PBsview[j]

    free(bytes_buffer)
    fclose(fi)

    return feature_metadata, data, metadata



import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from typing import List, Dict, Union

import anndata
from pegasusio import UnimodalData
from .views import INDEX, CINDEX, _parse_index, UnimodalDataView


class VDJDataView(UnimodalDataView):
    def __init__(self, vdjdata: "VDJData", barcode_index: CINDEX, feature_index: CINDEX, cur_matrix: str):
        super().__init__(vdjdata, barcode_index, feature_index, cur_matrix, obj_name = "VDJData")
        for keyword in self.parent._uns_keywords:
            self.metadata[keyword] = self.parent.metadata[keyword]
        self.obs # must call .obs in order to initialize self.barcode_metadata
        for chain in self.parent._features[self.metadata["modality"]]:
            keyword = "n" + chain
            if keyword in self.barcode_metadata:
                self.barcode_metadata.drop(columns = keyword, inplace = True)

            pos_arr = []
            for pos, vname in enumerate(self.var_names):
                if vname.startswith(chain):
                    pos_arr.append(pos)
            if len(pos_arr) > 0:
                self.barcode_metadata["n" + chain] = self.X[:, pos_arr].getnnz(axis = 1)

    def __repr__(self) -> str:
        return super().__repr__({"obs": str(list(self.barcode_metadata))[1:-1]})

    def __getitem__(self, index: INDEX) -> "VDJDataView":
        barcode_index, feature_index = _parse_index(self, index)
        return VDJDataView(self.parent, barcode_index, feature_index, self._cur_matrix)

    def get_chain(self, chain: str) -> pd.DataFrame:
        if chain not in self.var_names:
            raise ValueError(f"Chain '{chain}' is unknown!")

        data = {}
        fpos = self.var_names.get_loc(chain)
        for keyword in self.parent._matrix_keywords:
            uns_key = "_" + keyword
            self.select_matrix(keyword)
            if uns_key in self.uns:
                idx = self.X[:, fpos].toarray().ravel()
                data[keyword] = self.uns[uns_key][idx]
            else:
                data[keyword] = self.X[:, fpos].toarray().ravel()
        df = pd.DataFrame(data = data, index = self.obs_names, columns = self.parent._matrix_keywords)

        return df



class VDJData(UnimodalData):
    _matrix_keywords = ["high_confidence", "length", "reads", "umis", "v_gene", "d_gene", "j_gene", "c_gene", "cdr3", "cdr3_nt"]
    _uns_keywords = ["_v_gene", "_d_gene", "_j_gene", "_c_gene", "_cdr3", "_cdr3_nt"]
    _features = {"tcr": ["TRA", "TRB", "TRD", "TRG", "Multi"], "bcr": ["IGK", "IGL", "IGH", "Multi"]}
    _n_contigs = 10


    def __init__(
        self,
        barcode_metadata: Union[dict, pd.DataFrame],
        feature_metadata: Union[dict, pd.DataFrame],
        matrices: Dict[str, csr_matrix],
        metadata: dict,
        barcode_multiarrays: Dict[str, np.ndarray] = None,
        feature_multiarrays: Dict[str, np.ndarray] = None,
        cur_matrix: str = "umis",
    ) -> None:
        assert metadata["modality"] in {"tcr", "bcr"}
        super().__init__(barcode_metadata, feature_metadata, matrices, metadata, barcode_multiarrays, feature_multiarrays, cur_matrix)
        for chain in VDJData._features[self.uns["modality"]]:
            pos = self.var_names.get_loc(chain)
            self.obs["n" + chain] = self.X[:, pos:pos+VDJData._n_contigs].getnnz(axis = 1)


    def get_chain(self, chain: str) -> pd.DataFrame:
        if chain not in self.var_names:
            raise ValueError("Chain '{chain}' is unknown!")

        data = {}
        fpos = self.var_names.get_loc(chain)
        for keyword in VDJData._matrix_keywords:
            uns_key = "_" + keyword
            if uns_key in self.uns:
                idx = self.matrices[keyword][:, fpos].toarray().ravel()
                data[keyword] = self.uns[uns_key][idx]
            else:
                data[keyword] = self.matrices[keyword][:, fpos].toarray().ravel()
        df = pd.DataFrame(data = data, index = self.obs_names, columns = VDJData._matrix_keywords)

        return df


    def construct_clonotype(self, min_umis: int = 2) -> None:
        if self.get_modality() == "tcr":
            df1 = self.get_chain("TRA")
            df2 = self.get_chain("TRB")
            idx = (df1["umis"] >= min_umis) & (df2["umis"] >= min_umis)
            self.obs["has_clonotype"] = idx
            self.obs["clonotype"] = ""
            self.obs.loc[idx, "clonotype"] = df1.loc[idx, "cdr3"] + "|" + df2.loc[idx, "cdr3"]

            df3 = self.get_chain("TRD")
            df4 = self.get_chain("TRG")
            idx2 = (df3["umis"] >= min_umis) & (df4["umis"] >= min_umis) & (~idx)
            if idx2.sum() > 0:
                self.obs.loc[idx2, "has_clonotype"] = True
                self.obs.loc[idx2, "clonotype"] = df3.loc[idx2, "cdr3"] + "|" + df4.loc[idx2, "cdr3"]
        else:
            assert self.get_modality() == "bcr"
            df1 = self.get_chain("IGH")
            df2 = self.get_chain("IGK")
            df3 = self.get_chain("IGL")
            idx1 = df1["umis"] >= min_umis
            idx2 = (df2["umis"] >= df3["umis"]) & (df2["umis"] >= min_umis)
            idx3 = (df2["umis"] < df3["umis"]) & (df3["umis"] >= min_umis)
            idxkh = idx1 & idx2
            idxlh = idx1 & idx3
            self.obs["has_clonotype"] = idxkh | idxlh
            self.obs["clonotype"] = ""
            self.obs.loc[idxkh, "clonotype"] = df2.loc[idxkh, "cdr3"] + "|" + df1.loc[idxkh, "cdr3"]
            self.obs.loc[idxlh, "clonotype"] = df3.loc[idxlh, "cdr3"] + "|" + df1.loc[idxlh, "cdr3"]


    def from_anndata(self, data: anndata.AnnData, genome: str = None, modality: str = None) -> None:
        raise ValueError("Cannot convert an AnnData object to a VDJData object!")

    
    def to_anndata(self) -> anndata.AnnData:
        raise ValueError("Cannot convert a VDJData object ot an AnnData object!")


    def __getitem__(self, index: INDEX) -> VDJDataView:
        barcode_index, feature_index = _parse_index(self, index)
        return VDJDataView(self, barcode_index, feature_index, self._cur_matrix)



import os
import numpy as np
import pandas as pd

from pegasusio import NanostringData, MultimodalData

import logging
logger = logging.getLogger(__name__)



def load_nanostring_files(input_matrix: str, segment_file: str, annotation_file: str = None, genome: str = None) -> MultimodalData:
    """Load Nanostring GeoMx input files.

    Parameters
    ----------

    input_matrix : `str`
        Input Q3 normalized data matrix.
    segment_file: `str`
        Segment file containing segmentation information for each ROI. If segment_file == 'protein', load GeoMx protein results from nCounter.
    annotation_file: `str`, optional (default None)
        An optional annotation file providing tissue type information etc.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead.

    Returns
    -------

    A MultimodalData object containing a (genome, CytoData) pair.

    Examples
    --------
    >>> io.load_fcs_file('example.fcs', genome = 'GRCh38')
    """
    is_protein = segment_file == "protein"

    if not os.path.isfile(input_matrix):
        raise FileNotFoundError(f"File {input_matrix} does not exist!")

    genome = "unknown" if genome is None else genome
    metadata = {"genome": genome, "modality": "nanostring"}
    barcode_multiarrays = None

    if not is_protein:
        df = pd.read_csv(input_matrix, sep = '\t', header = 0, index_col = 0)

        barcodekey = pd.Index([x.replace('.', '-') for x in df.columns.values])
        barcode_metadata = {"barcodekey": barcodekey.values} # I guess the original matrix is processed in R because '-' -> '.'.
        feature_metadata = {"featurekey": df.index.values}
        matrices = {"Q3Norm": np.transpose(df.values)} # float64, do we need to convert it to float32?
        cur_matrix = "Q3Norm"

        if not os.path.isfile(segment_file):
            raise FileNotFoundError(f"File {segment_file} does not exist!")
        df = pd.read_csv(segment_file, sep = '\t', header = 0, index_col = 0)

        idx = barcodekey.isin(df.index)
        if idx.sum() < barcodekey.size:
            logger.warning(f"Cannot find {barcodekey[~idx]} from the segment property file! Number of AOIs reduces to {idx.sum()}.")
            barcodekey = barcodekey[idx]
            barcode_metadata["barcodekey"] = barcodekey.values
        if idx.sum() < df.shape[0]:
            logger.warning(f"Sample IDs {','.join(x for x in df.index[~df.index.isin(barcodekey)])} from the segment property file are not located in the matrix file!")
        df = df.reindex(barcodekey)

        for key in ["primer plate well", "slide name", "scan name", "panel", "segment", "aoi"]:
            if key in df.columns:
                df.loc[df[key].isna(), key] = "None"
                barcode_metadata[key] = df[key].values
        if "roi" in df.columns:
            rois = df["roi"].copy()
            rois[rois.isna()] = -1.0
            rois = rois.astype(int).astype(str)
            rois[rois == "-1"] = "None"
            barcode_metadata["roi"] = rois.values
        for key in ["area", "SequencingSaturation"]:
            if key in df.columns:
                df.loc[df[key].isna(), key] = 0.0
                barcode_metadata[key] = df[key].values.astype(np.float32)
        for key in ["RawReads", "TrimmedReads", "StitchedReads", "AlignedReads", "DeduplicatedReads"]:
            if key in df.columns:
                df.loc[df[key].isna(), key] = 0
                barcode_metadata[key] = df[key].values.astype(np.int32)
    else:
        df = pd.read_csv(input_matrix, sep = ',', header = None, index_col = 0)

        barcodekey = pd.Index(df.loc["Segment displayed name", 4:].values)
        barcode_metadata = {"barcodekey": barcodekey.values,
                            "segment": pd.Categorical(df.loc["Segment (Name/ Label)", 4:].values),
                            "AOI surface area": df.loc["AOI surface area", 4:].values.astype(np.float64),
                            "AOI nuclei count": df.loc["AOI nuclei count", 4:].values.astype(np.int32)
                           }
        probe_pos = df.index.get_loc("#Probe Group") + 1
        df_probes = df.iloc[probe_pos:]
        rawmat = np.transpose(df_probes.loc[:, 4:].values.astype(np.float64))
        idx_signal = df_probes[2] == "Endogenous"
        idx_control = df_probes[2] == "Control"
        idx_negative = df_probes[2] == "Negative"

        feature_metadata = {"featurekey": df_probes.loc[idx_signal, 3].values, "ProbeAnnotation": df_probes.index[idx_signal].values}
        matrices = {"RawData": rawmat[:, idx_signal]} # float64
        metadata["control_names"] = df_probes.loc[idx_control, 3].values
        metadata["negative_names"] = df_probes.loc[idx_negative, 3].values
        barcode_multiarrays = {"controls": rawmat[:, idx_control], "negatives": rawmat[:, idx_negative]}

        cur_matrix = "RawData"

    if annotation_file is not None:
        if not os.path.isfile(annotation_file):
            raise FileNotFoundError(f"File {annotation_file} does not exist!")
        df = pd.read_csv(annotation_file, sep = '\t', header = 0, index_col = 0)
        assert barcodekey.isin(df.index).sum() == barcodekey.size
        df = df.reindex(barcodekey)
        for key in df.columns:
            barcode_metadata[key] = pd.Categorical(df[key].values)


    nanodata = NanostringData(barcode_metadata, feature_metadata, matrices, metadata, barcode_multiarrays = barcode_multiarrays, cur_matrix = cur_matrix)
    data = MultimodalData(nanodata)

    return data



import os
import numpy as np
import pandas as pd

from pegasusio import CytoData, MultimodalData


def load_fcs_file(input_fcs: str, genome: str = None) -> MultimodalData:
    """Load Cyto data from a FCS file, support v2.0, v3.0 and v3.1.

    Parameters
    ----------

    input_fcs : `str`
        The FCS file.
    genome : `str`, optional (default None)
        The genome reference. If None, use "unknown" instead.

    Returns
    -------

    A MultimodalData object containing a (genome, CytoData) pair.

    Examples
    --------
    >>> io.load_fcs_file('example.fcs', genome = 'GRCh38')
    """
    try:
        from pegasusio.cylib.io import read_fcs
    except ModuleNotFoundError:
        print("No module named 'pegasusio.cylib.io'")

    if not os.path.isfile(input_fcs):
        raise FileNotFoundError(f"File {input_fcs} does not exist!")
    feature_metadata, matrix, metadata = read_fcs(input_fcs)
    barcode_metadata = {"barcodekey": [f"event{i}" for i in range(1, matrix.shape[0] + 1)]}
    genome = "unknown" if genome is None else genome
    metadata["genome"] = genome
    metadata["modality"] = "cyto"

    cytodata = CytoData(barcode_metadata, feature_metadata, {"raw.data": matrix}, metadata)
    data = MultimodalData(cytodata)

    return data



import numpy as np
import pandas as pd

from pegasusio import UnimodalData

import logging
logger = logging.getLogger(__name__)



class DictWithDefault:
    ### Used for parsing mito prefix
    def __init__(self, string: str):
        self.mapping = {}
        self.default = None

        if string is not None:
            fields = string.split(',')
            for field in fields:
                if field.find(':') >= 0:
                    key, value = field.split(':')
                    self.mapping[key] = value
                else:
                    self.default = field

    def get(self, key: str) -> str:
        return self.mapping.get(key, self.default)



def calc_qc_filters(
    unidata: UnimodalData,
    select_singlets: bool = False,
    remap_string: str = None,
    subset_string: str = None,
    min_genes: int = None,
    max_genes: int = None,
    min_umis: int = None,
    max_umis: int = None,
    mito_prefix: str = None,
    percent_mito: float = None
) -> None:
    """Calculate Quality Control (QC) metrics and mark barcodes based on the combination of QC metrics.

    Parameters
    ----------
    unidata: ``UnimodalData``
       Unimodal data matrix with rows for cells and columns for genes.
    select_singlets: ``bool``, optional, default ``False``
        If select only singlets.
    remap_string: ``str``, optional, default ``None``
        Remap singlet names using <remap_string>, where <remap_string> takes the format "new_name_i:old_name_1,old_name_2;new_name_ii:old_name_3;...". For example, if we hashed 5 libraries from 3 samples sample1_lib1, sample1_lib2, sample2_lib1, sample2_lib2 and sample3, we can remap them to 3 samples using this string: "sample1:sample1_lib1,sample1_lib2;sample2:sample2_lib1,sample2_lib2". In this way, the new singlet names will be in metadata field with key 'assignment', while the old names will be kept in metadata field with key 'assignment.orig'.
    subset_string: ``str``, optional, default ``None``
        If select singlets, only select singlets in the <subset_string>, which takes the format "name1,name2,...". Note that if --remap-singlets is specified, subsetting happens after remapping. For example, we can only select singlets from sampe 1 and 3 using "sample1,sample3".
    min_genes: ``int``, optional, default: None
       Only keep cells with at least ``min_genes`` genes.
    max_genes: ``int``, optional, default: None
       Only keep cells with less than ``max_genes`` genes.
    min_umis: ``int``, optional, default: None
       Only keep cells with at least ``min_umis`` UMIs.
    max_umis: ``int``, optional, default: None
       Only keep cells with less than ``max_umis`` UMIs.
    mito_prefix: ``str``, optional, default: None
       Prefix for mitochondrial genes.
    percent_mito: ``float``, optional, default: None
       Only keep cells with percent mitochondrial genes less than ``percent_mito`` % of total counts. Only when both mito_prefix and percent_mito set, the mitochondrial filter will be triggered.

    Returns
    -------
    ``None``

    Update ``unidata.obs``:

        * ``n_genes``: Total number of genes for each cell.
        * ``n_counts``: Total number of counts for each cell.
        * ``percent_mito``: Percent of mitochondrial genes for each cell.
        * ``passed_qc``: Boolean type indicating if a cell passes the QC process based on the QC metrics.
        * ``demux_type``: this column might be deleted if select_singlets is on.

    Examples
    --------
    >>> calc_qc_filters(unidata, min_umis = 500, select_singlets = True)
    """
    assert unidata.uns["modality"] == "rna"

    filters = []

    if select_singlets and ("demux_type" in unidata.obs):
        if remap_string is not None:
            if "assignment" not in unidata.obs:
                raise ValueError("No assignment field detected!")
            unidata.obs["assignment.orig"] = unidata.obs["assignment"]

            remap = {}
            tokens = remap_string.split(";")
            for token in tokens:
                new_key, old_str = token.split(":")
                old_keys = old_str.split(",")
                for key in old_keys:
                    remap[key] = new_key

            unidata.obs["assignment"] = pd.Categorical(unidata.obs["assignment"].apply(lambda x: remap[x] if x in remap else x))
            logger.info("Singlets are remapped.")


        if subset_string is None:
            filters.append(unidata.obs["demux_type"] == "singlet")
        else:
            if "assignment" not in unidata.obs:
                raise ValueError("No assignment field detected!")

            subset = np.array(subset_string.split(","))
            filters.append(np.isin(unidata.obs["assignment"], subset))

        unidata.uns["__del_demux_type"] = True

    if "n_genes" not in unidata.obs:
        unidata.obs["n_genes"] = unidata.X.getnnz(axis=1)

    if "n_counts" not in unidata.obs:
        unidata.obs["n_counts"] = unidata.X.sum(axis=1).A1

    if min_genes is not None:
        filters.append(unidata.obs["n_genes"] >= min_genes)
    if max_genes is not None:
        filters.append(unidata.obs["n_genes"] < max_genes)
    if min_umis is not None:
        filters.append(unidata.obs["n_counts"] >= min_umis)
    if max_umis is not None:
        filters.append(unidata.obs["n_counts"] < max_umis)

    if (mito_prefix is not None) and (percent_mito is not None):
        mito_genes = unidata.var_names.map(lambda x: x.startswith(mito_prefix)).values.nonzero()[0]
        unidata.obs["percent_mito"] = (
            unidata.X[:, mito_genes].sum(axis=1).A1
            / np.maximum(unidata.obs["n_counts"].values, 1.0)
        ) * 100
        filters.append(unidata.obs["percent_mito"] < percent_mito)

    if len(filters) > 0:
        selected = np.logical_and.reduce(filters)
        unidata.obs["passed_qc"] = selected
    else:
        unidata.obs["passed_qc"] = True


def apply_qc_filters(unidata: UnimodalData, uns_white_list: str = None):
    """ Apply QC filters to filter out low quality cells """
    if "passed_qc" in unidata.obs:
        prior_n = unidata.shape[0]
        unidata._inplace_subset_obs(unidata.obs["passed_qc"])

        cols = ["passed_qc"]
        if unidata.uns.get("__del_demux_type", False):
            cols.append("demux_type")
            if "assignment" in unidata.obs:
                # remove categories that contain no elements
                series = unidata.obs["assignment"].value_counts(sort = False)
                unidata.obs["assignment"] = pd.Categorical(unidata.obs["assignment"], categories = series[series > 0].index.astype(str))
            # del unidata.uns["__del_demux_type"]

        unidata.obs.drop(columns=cols, inplace=True)
        if len(unidata.obsm) > 0:
            unidata.obsm.clear()
        if len(unidata.varm) > 0:
            unidata.varm.clear()
        if uns_white_list is not None:
            white_list = set(uns_white_list.split(',') + ['genome', 'modality'])
            for key in list(unidata.uns):
                if key not in white_list:
                    del unidata.uns[key]
        logger.info(f"After filtration, {unidata.shape[0]} out of {prior_n} cell barcodes are kept in UnimodalData object {unidata.get_uid()}.")
